# sage.doctest: needs sage.libs.flint sage.libs.pari
"""
Eisenstein series
"""
# ****************************************************************************
#       Copyright (C) 2004-2006 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.functions import lcm
from sage.arith.misc import bernoulli, divisors, is_squarefree
from sage.misc.lazy_import import lazy_import
from sage.misc.timing import cputime
from sage.modular.arithgroup.congroup_gammaH import GammaH_class
from sage.modular.dirichlet import DirichletGroup
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.rings.power_series_ring import PowerSeriesRing
from sage.rings.rational_field import QQ

lazy_import('sage.modular.modform.eis_series_cython', ['eisenstein_series_poly', 'Ek_ZZ'])
lazy_import('sage.rings.number_field.number_field', 'CyclotomicField')


def eisenstein_series_qexp(k, prec=10, K=QQ, var='q', normalization='linear'):
    r"""
    Return the `q`-expansion of the normalized weight `k` Eisenstein series on
    `\SL_2(\ZZ)` to precision ``prec`` in the ring `K`. Three normalizations
    are available, depending on the parameter ``normalization``; the default
    normalization is the one for which the linear coefficient is 1.

    INPUT:

    - ``k`` -- an even positive integer

    - ``prec`` -- (default: 10) a nonnegative integer

    - ``K`` -- (default: `\QQ`) a ring

    - ``var`` -- (default: ``'q'``) variable name to use for `q`-expansion

    - ``normalization`` -- (default: ``'linear'``) normalization to use. If this
      is ``'linear'``, then the series will be normalized so that the linear
      term is 1. If it is ``'constant'``, the series will be normalized to have
      constant term 1. If it is ``'integral'``, then the series will be
      normalized to have integer coefficients and no common factor, and linear
      term that is positive. Note that ``'integral'`` will work over arbitrary
      base rings, while ``'linear'`` or ``'constant'`` will fail if the
      denominator (resp. numerator) of `B_k / 2k` is invertible.

    ALGORITHM:

        We know `E_k = \text{constant} + \sum_n \sigma_{k-1}(n) q^n`. So we
        compute all the `\sigma_{k-1}(n)` simultaneously, using the fact that
        `\sigma` is multiplicative.

    EXAMPLES::

        sage: eisenstein_series_qexp(2,5)
        -1/24 + q + 3*q^2 + 4*q^3 + 7*q^4 + O(q^5)
        sage: eisenstein_series_qexp(2,0)
        O(q^0)
        sage: eisenstein_series_qexp(2,5,GF(7))
        2 + q + 3*q^2 + 4*q^3 + O(q^5)
        sage: eisenstein_series_qexp(2,5,GF(7),var='T')
        2 + T + 3*T^2 + 4*T^3 + O(T^5)

    We illustrate the use of the ``normalization`` parameter::

        sage: eisenstein_series_qexp(12, 5, normalization='integral')
        691 + 65520*q + 134250480*q^2 + 11606736960*q^3 + 274945048560*q^4 + O(q^5)
        sage: eisenstein_series_qexp(12, 5, normalization='constant')
        1 + 65520/691*q + 134250480/691*q^2 + 11606736960/691*q^3 + 274945048560/691*q^4 + O(q^5)
        sage: eisenstein_series_qexp(12, 5, normalization='linear')
        691/65520 + q + 2049*q^2 + 177148*q^3 + 4196353*q^4 + O(q^5)
        sage: eisenstein_series_qexp(12, 50, K=GF(13), normalization='constant')
        1 + O(q^50)

    TESTS:

    Test that :issue:`5102` is fixed::

        sage: eisenstein_series_qexp(10, 30, GF(17))
        15 + q + 3*q^2 + 15*q^3 + 7*q^4 + 13*q^5 + 11*q^6 + 11*q^7 + 15*q^8 + 7*q^9 + 5*q^10 + 7*q^11 + 3*q^12 + 14*q^13 + 16*q^14 + 8*q^15 + 14*q^16 + q^17 + 4*q^18 + 3*q^19 + 6*q^20 + 12*q^21 + 4*q^22 + 12*q^23 + 4*q^24 + 4*q^25 + 8*q^26 + 14*q^27 + 9*q^28 + 6*q^29 + O(q^30)

    This shows that the bug reported at :issue:`8291` is fixed::

        sage: eisenstein_series_qexp(26, 10, GF(13))
        7 + q + 3*q^2 + 4*q^3 + 7*q^4 + 6*q^5 + 12*q^6 + 8*q^7 + 2*q^8 + O(q^10)

    We check that the function behaves properly over finite-characteristic base rings::

        sage: eisenstein_series_qexp(12, 5, K = Zmod(691), normalization='integral')
        566*q + 236*q^2 + 286*q^3 + 194*q^4 + O(q^5)
        sage: eisenstein_series_qexp(12, 5, K = Zmod(691), normalization='constant')
        Traceback (most recent call last):
        ...
        ValueError: The numerator of -B_k/(2*k) (=691) must be invertible in the ring Ring of integers modulo 691
        sage: eisenstein_series_qexp(12, 5, K = Zmod(691), normalization='linear')
        q + 667*q^2 + 252*q^3 + 601*q^4 + O(q^5)

        sage: eisenstein_series_qexp(12, 5, K = Zmod(2), normalization='integral')
        1 + O(q^5)
        sage: eisenstein_series_qexp(12, 5, K = Zmod(2), normalization='constant')
        1 + O(q^5)
        sage: eisenstein_series_qexp(12, 5, K = Zmod(2), normalization='linear')
        Traceback (most recent call last):
        ...
        ValueError: The denominator of -B_k/(2*k) (=65520) must be invertible in the ring Ring of integers modulo 2

    AUTHORS:

    - William Stein: original implementation

    - Craig Citro (2007-06-01): rewrote for massive speedup

    - Martin Raum (2009-08-02): port to cython for speedup

    - David Loeffler (2010-04-07): work around an integer overflow when `k` is large

    - David Loeffler (2012-03-15): add options for alternative normalizations
      (motivated by :issue:`12043`)
    """
    # we use this to prevent computation if it would fail anyway.
    if k <= 0 or k % 2 == 1:
        raise ValueError("k must be positive and even")

    a0 = - bernoulli(k) / (2*k)

    if normalization == 'linear':
        a0den = a0.denominator()
        try:
            a0fac = K(1/a0den)
        except ZeroDivisionError:
            raise ValueError("The denominator of -B_k/(2*k) (=%s) must be invertible in the ring %s" % (a0den, K))
    elif normalization == 'constant':
        a0num = a0.numerator()
        try:
            a0fac = K(1/a0num)
        except ZeroDivisionError:
            raise ValueError("The numerator of -B_k/(2*k) (=%s) must be invertible in the ring %s" % (a0num, K))
    elif normalization == 'integral':
        a0fac = None
    else:
        raise ValueError("Normalization (=%s) must be one of 'linear', 'constant', 'integral'" % normalization)

    R = PowerSeriesRing(K, var)
    if K == QQ and normalization == 'linear':
        ls = Ek_ZZ(k, prec)
        # The following is *dramatically* faster than doing the more natural
        # "R(ls)" would be:
        E = ZZ[var](ls, prec=prec, check=False).change_ring(QQ)
        if len(ls) > 0:
            E._unsafe_mutate(0, a0)
        return R(E, prec)
        # The following is an older slower alternative to the above three lines:
        # return a0fac*R(eisenstein_series_poly(k, prec).list(), prec=prec, check=False)
    # This used to work with check=False, but that can only be regarded as
    # an improbable lucky miracle. Enabling checking is a noticeable speed
    # regression; the morally right fix would be to expose FLINT's
    # fmpz_poly_to_nmod_poly command (at least for word-sized N).
    if a0fac is not None:
        return a0fac*R(eisenstein_series_poly(k, prec).list(), prec=prec, check=True)
    return R(eisenstein_series_poly(k, prec).list(), prec=prec, check=True)


def __common_minimal_basering(chi, psi):
    """
    Find the smallest basering over which chi and psi are valued, and
    return new chi and psi valued in that ring.

    EXAMPLES::

        sage: # needs sage.rings.number_field
        sage: sage.modular.modform.eis_series.__common_minimal_basering(DirichletGroup(1)[0], DirichletGroup(1)[0])
        (Dirichlet character modulo 1 of conductor 1,
         Dirichlet character modulo 1 of conductor 1)
        sage: sage.modular.modform.eis_series.__common_minimal_basering(DirichletGroup(3).0, DirichletGroup(5).0)
        (Dirichlet character modulo 3 of conductor 3 mapping 2 |--> -1,
         Dirichlet character modulo 5 of conductor 5 mapping 2 |--> zeta4)
        sage: sage.modular.modform.eis_series.__common_minimal_basering(DirichletGroup(12).0, DirichletGroup(36).0)
        (Dirichlet character modulo 12 of conductor 4 mapping 7 |--> -1, 5 |--> 1,
         Dirichlet character modulo 36 of conductor 4 mapping 19 |--> -1, 29 |--> 1)
    """
    chi = chi.minimize_base_ring()
    psi = psi.minimize_base_ring()
    n = lcm(chi.base_ring().zeta().multiplicative_order(),
                  psi.base_ring().zeta().multiplicative_order())
    if n <= 2:
        K = QQ
    else:
        K = CyclotomicField(n)
    chi = chi.change_ring(K)
    psi = psi.change_ring(K)
    return chi, psi


def __find_eisen_chars(character, k):
    r"""
    Find all triples `(\psi_1, \psi_2, t)` that give rise to an Eisenstein series of the given weight and character.

    EXAMPLES::

        sage: sage.modular.modform.eis_series.__find_eisen_chars(DirichletGroup(36).0, 4)
        []

        sage: pars =  sage.modular.modform.eis_series.__find_eisen_chars(DirichletGroup(36).0, 5)
        sage: [(x[0].values_on_gens(), x[1].values_on_gens(), x[2]) for x in pars]
        [((1, 1), (-1, 1), 1),
        ((1, 1), (-1, 1), 3),
        ((1, 1), (-1, 1), 9),
        ((1, -1), (-1, -1), 1),
        ((-1, 1), (1, 1), 1),
        ((-1, 1), (1, 1), 3),
        ((-1, 1), (1, 1), 9),
        ((-1, -1), (1, -1), 1)]
    """
    from sage.misc.verbose import verbose

    N = character.modulus()
    if character.is_trivial():
        if k % 2:
            return []
        char_inv = ~character
        V = [(character, char_inv, t) for t in divisors(N) if t > 1]
        if k != 2:
            V.insert(0, (character, char_inv, 1))
        if is_squarefree(N):
            return V
        # Now include all pairs (chi,chi^(-1)) such that cond(chi)^2 divides N:
        # TODO: Optimize -- this is presumably way too hard work below.
        G = DirichletGroup(N)
        for chi in G:
            if not chi.is_trivial():
                f = chi.conductor()
                if N % (f**2) == 0:
                    chi = chi.minimize_base_ring()
                    chi_inv = ~chi
                    for t in divisors(N//(f**2)):
                        V.insert(0, (chi, chi_inv, t))
        return V

    eps = character
    if eps(-1) != (-1)**k:
        return []
    eps = eps.maximize_base_ring()
    G = eps.parent()

    # Find all pairs chi, psi such that:
    #
    #  (1) cond(chi)*cond(psi) divides the level, and
    #
    #  (2) chi*psi == eps, where eps is the nebentypus character of self.
    #
    # See [Miyake, Modular Forms] Lemma 7.1.1.

    C = {}

    t0 = cputime()

    for e in G:
        m = Integer(e.conductor())
        if m in C:
            C[m].append(e)
        else:
            C[m] = [e]

    verbose("Enumeration with conductors.", t0)

    params = []
    for L in divisors(N):
        verbose("divisor %s" % L)
        if L not in C:
            continue
        GL = C[L]
        for R in divisors(N // L):
            if R not in C:
                continue
            GR = C[R]
            for chi in GL:
                for psi in GR:
                    if chi*psi == eps:
                        chi0, psi0 = __common_minimal_basering(chi, psi)
                        for t in divisors(N//(R*L)):
                            if k != 1 or (psi0, chi0, t) not in params:
                                params.append((chi0, psi0, t))
    return params


def __find_eisen_chars_gammaH(N, H, k):
    r"""
    Find all triples `(\psi_1, \psi_2, t)` that give rise to an Eisenstein series of weight `k` on
    `\Gamma_H(N)`.

    EXAMPLES::

        sage: pars =  sage.modular.modform.eis_series.__find_eisen_chars_gammaH(15, [2], 5)
        sage: [(x[0].values_on_gens(), x[1].values_on_gens(), x[2]) for x in pars]
        [((1, 1), (-1, -1), 1), ((-1, 1), (1, -1), 1), ((1, -1), (-1, 1), 1), ((-1, -1), (1, 1), 1)]
    """
    params = []
    for chi in DirichletGroup(N):
        if all(chi(h) == 1 for h in H):
            params += __find_eisen_chars(chi, k)
    return params


def __find_eisen_chars_gamma1(N, k):
    r"""
    Find all triples `(\psi_1, \psi_2, t)` that give rise to an Eisenstein series of weight `k` on
    `\Gamma_1(N)`.

    EXAMPLES::

        sage: pars = sage.modular.modform.eis_series.__find_eisen_chars_gamma1(12, 4)
        sage: [(x[0].values_on_gens(), x[1].values_on_gens(), x[2]) for x in pars]
        [((1, 1), (1, 1), 1),
        ((1, 1), (1, 1), 2),
        ((1, 1), (1, 1), 3),
        ((1, 1), (1, 1), 4),
        ((1, 1), (1, 1), 6),
        ((1, 1), (1, 1), 12),
        ((1, 1), (-1, -1), 1),
        ((-1, -1), (1, 1), 1),
        ((-1, 1), (1, -1), 1),
        ((1, -1), (-1, 1), 1)]

        sage: pars =  sage.modular.modform.eis_series.__find_eisen_chars_gamma1(12, 5)
        sage: [(x[0].values_on_gens(), x[1].values_on_gens(), x[2]) for x in pars]
        [((1, 1), (-1, 1), 1),
        ((1, 1), (-1, 1), 3),
        ((-1, 1), (1, 1), 1),
        ((-1, 1), (1, 1), 3),
        ((1, 1), (1, -1), 1),
        ((1, 1), (1, -1), 2),
        ((1, 1), (1, -1), 4),
        ((1, -1), (1, 1), 1),
        ((1, -1), (1, 1), 2),
        ((1, -1), (1, 1), 4)]
    """
    pairs = []
    s = (-1)**k
    G = DirichletGroup(N)
    E = list(G)
    parity = [c(-1) for c in E]
    for i in range(len(E)):
        for j in range(i, len(E)):
            if parity[i] * parity[j] == s and N % (E[i].conductor() * E[j].conductor()) == 0:
                chi, psi = __common_minimal_basering(E[i], E[j])
                if k != 1:
                    pairs.append((chi, psi))
                    if i != j:
                        pairs.append((psi, chi))
                else:
                    # if weight is 1 then (chi, psi) and (chi, psi) are the
                    # same form
                    if psi.is_trivial() and not chi.is_trivial():
                        # need to put the trivial character first to get the L-value right
                        pairs.append((psi, chi))
                    else:
                        pairs.append((chi, psi))

    triples = []
    for chi, psi in pairs:
        c_chi = chi.conductor()
        c_psi = psi.conductor()
        D = divisors(N // (c_chi * c_psi))
        if k == 2 and chi.is_trivial() and psi.is_trivial():
            D.remove(1)
        chi, psi = __common_minimal_basering(chi, psi)
        triples.extend((chi, psi, t) for t in D)

    return triples


def eisenstein_series_lseries(weight, prec=53, max_imaginary_part=0):
    r"""
    Return the `L`-series of the weight `2k` Eisenstein series `E_{2k}`
    on `\SL_2(\ZZ)`.

    This returns an interface to Pari for computing with the
    `L`-series of the Eisenstein series.

    See :class:`~sage.lfunctions.pari.lfun_eisenstein`.

    INPUT:

    - ``weight`` -- even integer

    - ``prec`` -- integer (bits precision)

    - ``max_imaginary_part`` -- real number (default: 0)

    OUTPUT: the `L`-series of the Eisenstein series. This can be
    evaluated at argument `s` and has methods like `derivative`, etc.

    EXAMPLES:

    We compute with the `L`-series of `E_{16}` and then `E_{20}`::

        sage: L = eisenstein_series_lseries(16)
        sage: L(1)
        -0.291657724743874
        sage: L.derivative(1)
        0.0756072194360656
        sage: L = eisenstein_series_lseries(20)
        sage: L(2)
        -5.02355351645998

    Now with higher precision::

        sage: L = eisenstein_series_lseries(20, prec=200)
        sage: L(2)
        -5.0235535164599797471968418348135050804419155747868718371029
    """
    # ref : https://arxiv.org/pdf/1904.00190 Example 5.3
    from sage.lfunctions.pari import lfun_eisenstein, LFunction
    L = LFunction(lfun_eisenstein(weight), prec=prec,
                  max_im=max_imaginary_part)
    L.rename(f'L-series associated to the Eisenstein series E{weight} '
             'on SL_2(Z)')
    return L


def compute_eisenstein_params(character, k):
    r"""
    Compute and return a list of all parameters `(\chi,\psi,t)` that
    define the Eisenstein series with given character and weight `k`.

    Only the parity of `k` is relevant (unless k = 1, which is a slightly different case).

    If ``character`` is an integer `N`, then the parameters for
    `\Gamma_1(N)` are computed instead.  Then the condition is that
    `\chi(-1)*\psi(-1) =(-1)^k`.

    If ``character`` is a list of integers, the parameters for `\Gamma_H(N)` are
    computed, where `H` is the subgroup of `(\ZZ/N\ZZ)^\times` generated by the
    integers in the given list.

    EXAMPLES::

        sage: sage.modular.modform.eis_series.compute_eisenstein_params(DirichletGroup(30)(1), 3)
        []

        sage: pars =  sage.modular.modform.eis_series.compute_eisenstein_params(DirichletGroup(30)(1), 4)
        sage: [(x[0].values_on_gens(), x[1].values_on_gens(), x[2]) for x in pars]
        [((1, 1), (1, 1), 1),
        ((1, 1), (1, 1), 2),
        ((1, 1), (1, 1), 3),
        ((1, 1), (1, 1), 5),
        ((1, 1), (1, 1), 6),
        ((1, 1), (1, 1), 10),
        ((1, 1), (1, 1), 15),
        ((1, 1), (1, 1), 30)]

        sage: pars = sage.modular.modform.eis_series.compute_eisenstein_params(15, 1)
        sage: [(x[0].values_on_gens(), x[1].values_on_gens(), x[2]) for x in pars]
        [((1, 1), (-1, 1), 1),
        ((1, 1), (-1, 1), 5),
        ((1, 1), (1, zeta4), 1),
        ((1, 1), (1, zeta4), 3),
        ((1, 1), (-1, -1), 1),
        ((1, 1), (1, -zeta4), 1),
        ((1, 1), (1, -zeta4), 3),
        ((-1, 1), (1, -1), 1)]

        sage: sage.modular.modform.eis_series.compute_eisenstein_params(DirichletGroup(15).0, 1)
        [(Dirichlet character modulo 15 of conductor 1 mapping 11 |--> 1, 7 |--> 1, Dirichlet character modulo 15 of conductor 3 mapping 11 |--> -1, 7 |--> 1, 1),
        (Dirichlet character modulo 15 of conductor 1 mapping 11 |--> 1, 7 |--> 1, Dirichlet character modulo 15 of conductor 3 mapping 11 |--> -1, 7 |--> 1, 5)]

        sage: len(sage.modular.modform.eis_series.compute_eisenstein_params(GammaH(15, [4]), 3))
        8
    """
    if isinstance(character, (int, Integer)):
        return __find_eisen_chars_gamma1(character, k)
    if isinstance(character, GammaH_class):
        return __find_eisen_chars_gammaH(character.level(), character._generators_for_H(), k)
    return __find_eisen_chars(character, k)



from sage.structure.element cimport AlgebraElement
from sage.rings.power_series_ring_element cimport PowerSeries

cdef class LaurentSeries(AlgebraElement):
    cdef PowerSeries __u
    cdef long __n

    cdef _normalize(self)
    cpdef _add_(self, other)
    cpdef _mul_(self, other)



"""
Quadratic forms overview

AUTHORS:

- Jon Hanke (2007-06-19)
- Anna Haensch (2010-07-01): Formatting and ReSTification
- Simon Brandhorst (2019-10-15): :meth:`quadratic_form_from_invariants`
"""

# ****************************************************************************
#       Copyright (C) 2007 William Stein and Jonathan Hanke
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from copy import deepcopy
from warnings import warn

from sage.arith.functions import lcm as LCM
from sage.arith.misc import GCD
from sage.categories.fields import Fields
from sage.categories.principal_ideal_domains import PrincipalIdealDomains
from sage.categories.rings import Rings
from sage.matrix.constructor import matrix
from sage.matrix.matrix_space import MatrixSpace
from sage.misc.functional import denominator, is_even
from sage.misc.lazy_import import lazy_import
from sage.modules.free_module_element import vector
from sage.quadratic_forms.quadratic_form__evaluate import (
    QFEvaluateMatrix,
    QFEvaluateVector,
)
from sage.rings.ideal import Ideal
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.multi_polynomial import MPolynomial
from sage.rings.polynomial.polynomial_element import Polynomial
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.structure.element import Matrix, Vector
from sage.structure.sage_object import SageObject


def quadratic_form_from_invariants(F, rk, det, P, sminus):
    r"""
    Return a rational quadratic form with given invariants.

    INPUT:

    - ``F`` -- the base field; currently only ``QQ`` is allowed
    - ``rk`` -- integer; the rank
    - ``det`` -- rational; the determinant
    - ``P`` -- list of primes where Cassel's Hasse invariant
      is negative
    - ``sminus`` -- integer; the number of negative eigenvalues
      of any Gram matrix

    OUTPUT: a quadratic form with the specified invariants

    Let `(a_1, \ldots, a_n)` be the Gram matrix of a regular quadratic space.
    Then Cassel's Hasse invariant is defined as

    .. MATH::

        \prod_{i<j} (a_i,a_j),

    where `(a_i,a_j)` denotes the Hilbert symbol.

    ALGORITHM:

    We follow [Kir2016]_.

    EXAMPLES::

        sage: P = [3,5]
        sage: q = quadratic_form_from_invariants(QQ,2,-15,P,1); q                       # needs sage.rings.padics
        Quadratic form in 2 variables over Rational Field with coefficients:
        [ 5 0 ]
        [ * -3 ]
        sage: all(q.hasse_invariant(p) == -1 for p in P)                                # needs sage.rings.padics
        True

    TESTS:

    This shows that :issue:`28955` is fixed::

        sage: quadratic_form_from_invariants(QQ,3,2,[2],2)                              # needs sage.rings.padics
        Quadratic form in 3 variables over Rational Field with coefficients:
        [ -1 0 0 ]
        [ * 1 0 ]
        [ * * -2 ]

        sage: quadratic_form_from_invariants(QQ,4,2,[2],4)                              # needs sage.rings.padics
        Traceback (most recent call last):
        ...
        ValueError: invariants do not define a rational quadratic form
    """
    from sage.arith.misc import hilbert_symbol
    # normalize input
    if F != QQ:
        raise NotImplementedError('base field must be QQ. If you want this over any field, implement weak approximation.')
    P = [ZZ(p) for p in P]
    rk = ZZ(rk)
    d = QQ(det).squarefree_part()
    sminus = ZZ(sminus)
    # check if the invariants define a global quadratic form
    if d.sign() != (-1)**sminus:
        raise ValueError("invariants do not define a rational quadratic form")
    if rk == 1 and len(P) != 0:
        raise ValueError("invariants do not define a rational quadratic form")
    if rk == 2:
        for p in P:
            if QQ(-d).is_padic_square(p):
                raise ValueError("invariants do not define a rational quadratic form")
    f = 0
    if sminus % 4 in (2, 3):
        f = 1
    if (f + len(P)) % 2:
        raise ValueError("invariants do not define a rational quadratic form")
    D = []
    while rk >= 2:
        if rk >= 4:
            if sminus > 0:
                a = ZZ(-1)
            else:
                a = ZZ.one()
        elif rk == 3:
            Pprime = [p for p in P if hilbert_symbol(-1, -d, p) == 1]
            Pprime += [p for p in (2 * d).prime_divisors()
                       if hilbert_symbol(-1, -d, p) == -1 and p not in P]
            if sminus > 0:
                a = ZZ(-1)
            else:
                a = ZZ.one()
            for p in Pprime:
                if d.valuation(p) % 2 == 0:
                    a *= p
            assert all((a * d).valuation(p) % 2 == 1 for p in Pprime)
        elif rk == 2:
            S = P
            if sminus == 2:
                S += [-1]
            a = QQ.hilbert_symbol_negative_at_S(S, -d)
            a = ZZ(a)
        P = ([p for p in P if hilbert_symbol(a, -d, p) == 1]
             + [p for p in (2 * a * d).prime_divisors()
                if hilbert_symbol(a, -d, p) == -1 and p not in P])
        sminus = max(0, sminus - 1)
        rk = rk - 1
        d = a * d
        D.append(a.squarefree_part())
    d = d.squarefree_part()
    D.append(d)
    return DiagonalQuadraticForm(QQ, D)


class QuadraticForm(SageObject):
    r"""
    The ``QuadraticForm`` class represents a quadratic form in `n` variables with
    coefficients in the ring `R`.

    INPUT:

    The constructor may be called in any of the following ways.

    #. ``QuadraticForm(R, n, entries)``, where

       - ``R`` -- ring for which the quadratic form is defined
       - ``n`` -- integer `\geq 0`
       - ``entries`` -- list of `n(n+1)/2` coefficients of the quadratic form
         in `R` (given lexicographically, or equivalently, by rows of the
         matrix)

    #. ``QuadraticForm(p)``, where

       - ``p`` -- a homogeneous polynomial of degree `2`

    #. ``QuadraticForm(R, n)``, where

       - ``R`` -- a ring
       - ``n`` -- a symmetric `n \times n` matrix with even diagonal (relative to
         `R`)

    #. ``QuadraticForm(R)``, where

       - ``R`` -- a symmetric `n \times n` matrix with even diagonal (relative to
         its base ring)

    If the keyword argument ``unsafe_initialize`` is True, then the subsequent
    fields may by used to force the external initialization of various fields
    of the quadratic form. Currently the only fields which can be set are:

    - ``number_of_automorphisms``
    - ``determinant``

    OUTPUT: quadratic form

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 3, [1,2,3,4,5,6]); Q
        Quadratic form in 3 variables over Integer Ring with coefficients:
        [ 1 2 3 ]
        [ * 4 5 ]
        [ * * 6 ]

    ::

        sage: Q = QuadraticForm(QQ, 3, [1,2,3,4/3,5,6]); Q
        Quadratic form in 3 variables over Rational Field with coefficients:
        [ 1 2 3 ]
        [ * 4/3 5 ]
        [ * * 6 ]
        sage: Q[0,0]
        1
        sage: Q[0,0].parent()
        Rational Field

    ::

        sage: Q = QuadraticForm(QQ, 7, range(28)); Q
        Quadratic form in 7 variables over Rational Field with coefficients:
        [ 0 1 2 3 4 5 6 ]
        [ * 7 8 9 10 11 12 ]
        [ * * 13 14 15 16 17 ]
        [ * * * 18 19 20 21 ]
        [ * * * * 22 23 24 ]
        [ * * * * * 25 26 ]
        [ * * * * * * 27 ]

    ::

        sage: Q = QuadraticForm(QQ, 2, range(1,4))
        sage: A = Matrix(ZZ, 2, 2, [-1,0,0,1])
        sage: Q(A)
        Quadratic form in 2 variables over Rational Field with coefficients:
        [ 1 -2 ]
        [ * 3 ]

    ::

        sage: m = matrix(2, 2, [1,2,3,4])
        sage: m + m.transpose()
        [2 5]
        [5 8]
        sage: QuadraticForm(m + m.transpose())
        Quadratic form in 2 variables over Integer Ring with coefficients:
        [ 1 5 ]
        [ * 4 ]

    ::

        sage: P.<x,y,z> = QQ[]
        sage: p = x^2 + 2*x*y + x*z/2 + y^2 + y*z/3
        sage: QuadraticForm(p)
        Quadratic form in 3 variables over Rational Field with coefficients:
        [ 1 2 1/2 ]
        [ * 1 1/3 ]
        [ * * 0 ]

    ::

        sage: QuadraticForm(ZZ, m + m.transpose())
        Quadratic form in 2 variables over Integer Ring with coefficients:
        [ 1 5 ]
        [ * 4 ]

    ::

        sage: QuadraticForm(QQ, m + m.transpose())
        Quadratic form in 2 variables over Rational Field with coefficients:
        [ 1 5 ]
        [ * 4 ]
    """

    # Import specialized methods:
    # ---------------------------

    # Routines to compute the p-adic local normal form
    lazy_import("sage.quadratic_forms.quadratic_form__local_normal_form", [
        "find_entry_with_minimal_scale_at_prime",
        "local_normal_form",
        "jordan_blocks_by_scale_and_unimodular",
        "jordan_blocks_in_unimodular_list_by_scale_power"
    ])

    # Routines to compute local densities by counting solutions of various types
    from sage.quadratic_forms.quadratic_form__count_local_2 import (
        count_congruence_solutions,
        count_congruence_solutions__bad_type,
        count_congruence_solutions__bad_type_I,
        count_congruence_solutions__bad_type_II,
        count_congruence_solutions__good_type,
        count_congruence_solutions__zero_type,
        count_congruence_solutions_as_vector,
    )

    # Routines to compute local densities by the reduction procedure
    from sage.quadratic_forms.quadratic_form__local_density_congruence import (
        count_modp_solutions__by_Gauss_sum,
        local_bad_density_congruence,
        local_badI_density_congruence,
        local_badII_density_congruence,
        local_density_congruence,
        local_good_density_congruence,
        local_good_density_congruence_even,
        local_good_density_congruence_odd,
        local_primitive_density_congruence,
        local_zero_density_congruence,
    )

    # Routines to compute p-adic field invariants
    from sage.quadratic_forms.quadratic_form__local_field_invariants import (
        _rational_diagonal_form_and_transformation,
        anisotropic_primes,
        compute_definiteness,
        compute_definiteness_string_by_determinants,
        hasse_invariant,
        hasse_invariant__OMeara,
        is_anisotropic,
        is_definite,
        is_hyperbolic,
        is_indefinite,
        is_isotropic,
        is_negative_definite,
        is_positive_definite,
        rational_diagonal_form,
        signature,
        signature_vector,
    )

    # Routines to perform elementary variable substitutions
    from sage.quadratic_forms.quadratic_form__variable_substitutions import (
        add_symmetric,
        divide_variable,
        elementary_substitution,
        extract_variables,
        multiply_variable,
        scale_by_factor,
        swap_variables,
    )

    # Routines to be called by the user to compute local densities
    lazy_import('sage.quadratic_forms.quadratic_form__local_density_interfaces', [
        'local_density',
        'local_primitive_density'
    ])

    # Routines for computing with ternary forms
    from sage.quadratic_forms.quadratic_form__ternary_Tornaria import (
        adjoint,
        antiadjoint,
        basiclemma,
        basiclemmavec,
        clifford_conductor,
        clifford_invariant,
        content,
        delta,
        disc,
        discrec,
        hasse_conductor,
        is_adjoint,
        is_zero,
        is_zero_nonsingular,
        is_zero_singular,
        level__Tornaria,
        lll,
        omega,
        reciprocal,
        representation_number_list,
        representation_vector_list,
        xi,
        xi_rec,
    )

    # Routines to compute the theta function
    from sage.quadratic_forms.quadratic_form__theta import (
        theta_by_cholesky,
        theta_by_pari,
        theta_series,
        theta_series_degree_2,
    )

    # Routines to compute the product of all local densities
    lazy_import('sage.quadratic_forms.quadratic_form__siegel_product', [
        'siegel_product'
    ])

    # Routines to compute p-neighbors
    from sage.quadratic_forms.quadratic_form__neighbors import (
        find_p_neighbor_from_vec,
        find_primitive_p_divisible_vector__next,
        find_primitive_p_divisible_vector__random,
        neighbor_iteration,
        orbits_lines_mod_p,
    )

    # Routines to reduce a given quadratic form
    from sage.quadratic_forms.quadratic_form__reduction_theory import (
        minkowski_reduction,
        minkowski_reduction_for_4vars__SP,
        reduced_binary_form,
        reduced_binary_form1,
        reduced_ternary_form__Dickson,
    )
    # Wrappers for Conway-Sloane genus routines (in ./genera/)
    lazy_import('sage.quadratic_forms.quadratic_form__genus', [
        'global_genus_symbol',
        'local_genus_symbol',
        'CS_genus_symbol_list'
    ])

    # Routines to compute local masses for ZZ.
    lazy_import('sage.quadratic_forms.quadratic_form__mass', [
        'shimura_mass__maximal',
        'GHY_mass__maximal'
    ])
    lazy_import('sage.quadratic_forms.quadratic_form__mass__Siegel_densities', [
        'mass__by_Siegel_densities',
        'Pall_mass_density_at_odd_prime',
        'Watson_mass_at_2',
        'Kitaoka_mass_at_2',
        'mass_at_two_by_counting_mod_power'
    ])
    lazy_import('sage.quadratic_forms.quadratic_form__mass__Conway_Sloane_masses', [
        'parity',
        'is_even',
        'is_odd',
        'conway_species_list_at_odd_prime',
        'conway_species_list_at_2',
        'conway_octane_of_this_unimodular_Jordan_block_at_2',
        'conway_diagonal_factor',
        'conway_cross_product_doubled_power',
        'conway_type_factor',
        'conway_p_mass',
        'conway_standard_p_mass',
        'conway_standard_mass',
        'conway_mass'
        #            conway_generic_mass, \
        #            conway_p_mass_adjustment
    ])

    # Routines to check local representability of numbers
    lazy_import('sage.quadratic_forms.quadratic_form__local_representation_conditions', [
        'local_representation_conditions',
        'is_locally_universal_at_prime',
        'is_locally_universal_at_all_primes',
        'is_locally_universal_at_all_places',
        'is_locally_represented_number_at_place',
        'is_locally_represented_number'
    ])

    # Routines to make a split local covering of the given quadratic form.
    from sage.quadratic_forms.quadratic_form__split_local_covering import (
        cholesky_decomposition,
        complementary_subform_to_vector,
        split_local_cover,
        vectors_by_length,
    )

    # Routines to make automorphisms of the given quadratic form.
    lazy_import('sage.quadratic_forms.quadratic_form__automorphisms', [
        'basis_of_short_vectors',
        'short_vector_list_up_to_length',
        'short_primitive_vector_list_up_to_length',
        '_compute_automorphisms',
        'automorphism_group',
        'automorphisms',
        'number_of_automorphisms',
        'set_number_of_automorphisms'
    ])

    # Routines to test the local and global equivalence/isometry of two quadratic forms.
    from sage.quadratic_forms.quadratic_form__equivalence_testing import (
        has_equivalent_Jordan_decomposition_at_prime,
        is_globally_equivalent_to,
        is_locally_equivalent_to,
        is_rationally_isometric,
    )

    # Routines for solving equations of the form Q(x) = c.
    lazy_import('sage.quadratic_forms.qfsolve', [
        'solve'
    ])

    # Genus
    lazy_import("sage.quadratic_forms.genera.genus", ["genera"])

    def __init__(
        self,
        R,
        n=None,
        entries=None,
        unsafe_initialization=False,
        number_of_automorphisms=None,
        determinant=None,
    ) -> None:
        """
        EXAMPLES::

            sage: s = QuadraticForm(ZZ, 4, range(10))
            sage: s.dim()
            4

            sage: P.<x,y,z> = QQ[]
            sage: p = x^2 + y^2 + 2*x*z
            sage: QuadraticForm(p)
            Quadratic form in 3 variables over Rational Field with coefficients:
            [ 1 0 2 ]
            [ * 1 0 ]
            [ * * 0 ]
            sage: z = P.zero()
            sage: QuadraticForm(z)
            Quadratic form in 3 variables over Rational Field with coefficients:
            [ 0 0 0 ]
            [ * 0 0 ]
            [ * * 0 ]
            sage: q = x^2 + 3*y - z
            sage: QuadraticForm(q)
            Traceback (most recent call last):
            ...
            ValueError: polynomial is neither zero nor homogeneous of degree 2

        TESTS::

            sage: s == loads(dumps(s))
            True
            sage: QuadraticForm(ZZ, -1)
            Traceback (most recent call last):
            ...
            ValueError: the size must be a nonnegative integer, not -1

            sage: x = polygen(ZZ, 'x')
            sage: QuadraticForm(x**2)
            Quadratic form in 1 variables over Integer Ring with coefficients:
            [ 1 ]

            sage: QuadraticForm(1)
            Traceback (most recent call last):
            ....
            TypeError: wrong input for QuadraticForm
        """
        # Deal with:  QuadraticForm(ring, matrix)
        matrix_init_flag = False
        if R in Rings():
            if isinstance(n, Matrix):
                # Test if n is symmetric and has even diagonal
                if not self._is_even_symmetric_matrix_(n, R):
                    raise TypeError("the matrix is not a symmetric with even diagonal defined over R")

                # Rename the matrix and ring
                M = n
                M_ring = R
                matrix_init_flag = True

        elif isinstance(R, Matrix):
            M = R

            # Test if R is symmetric and has even diagonal
            if not self._is_even_symmetric_matrix_(M):
                raise TypeError("the matrix is not a symmetric with even diagonal")

            M_ring = M.base_ring()
            matrix_init_flag = True

        elif isinstance(R, (Polynomial, MPolynomial)):
            p = R

            if not p.is_zero() and not (p.is_homogeneous() and p.degree() == 2):
                raise ValueError("polynomial is neither zero nor homogeneous of degree 2")

            P = p.parent()
            R, n = P.base_ring(), P.ngens()

            # Extract quadratic form coefficients
            entries = []
            if n == 0:
                exponents = []
            elif n == 1:
                exponents = [2]
            else:
                from sage.combinat.integer_lists.invlex import IntegerListsLex

                exponents = IntegerListsLex(2, length=n)
            for alpha in exponents:
                entries.append(p[alpha])

        else:
            raise TypeError('wrong input for QuadraticForm')

        # Perform the quadratic form initialization
        if matrix_init_flag:
            self.__n = ZZ(M.nrows())
            self.__base_ring = M_ring
            self.__coeffs = []
            for i in range(M.nrows()):
                for j in range(i, M.nrows()):
                    if i == j:
                        self.__coeffs += [M_ring(M[i, j] / 2)]
                    else:
                        self.__coeffs += [M_ring(M[i, j])]

            return

        # -----------------------------------------------------------

        # Verify the size of the matrix is an integer >= 0
        n = ZZ(n)
        if n < 0:
            raise ValueError(f"the size must be a nonnegative integer, not {n}")

        # Store the relevant variables
        N = n * (n + 1) // 2
        self.__n = n
        self.__base_ring = R
        self.__coeffs = [self.__base_ring.zero() for i in range(N)]

        # Check if entries is a list, tuple or iterator for the
        # current size, and if so, write the upper-triangular matrix
        if entries is not None:
            try:
                entries = list(entries)
            except TypeError:
                raise TypeError('entries must be an iterable')

            if len(entries) == N:
                for i in range(N):
                    self.__coeffs[i] = self.__base_ring(entries[i])
            else:
                raise TypeError(f"the entries {entries} must be a list of size n(n+1)/2")

        # -----------------------------------------------------------

        # Process possible forced initialization of various fields
        self._external_initialization_list = []
        if unsafe_initialization:

            # Set the number of automorphisms
            if number_of_automorphisms is not None:
                self.set_number_of_automorphisms(number_of_automorphisms)
                # self.__number_of_automorphisms = number_of_automorphisms
                # self.__external_initialization_list.append('number_of_automorphisms')

            # Set the determinant
            if determinant is not None:
                self.__det = determinant
                self._external_initialization_list.append('determinant')

    def list_external_initializations(self) -> list:
        """
        Return a list of the fields which were set externally at
        creation, and not created through the usual :class:`QuadraticForm`
        methods.

        These fields are as good as the external process
        that made them, and are thus not guaranteed to be correct.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,0,5])
            sage: Q.list_external_initializations()
            []

            sage: # needs sage.libs.pari
            sage: T = Q.theta_series()
            sage: Q.list_external_initializations()
            []
            sage: Q = QuadraticForm(ZZ, 2, [1,0,5], unsafe_initialization=False,
            ....:                   number_of_automorphisms=3, determinant=0)
            sage: Q.list_external_initializations()
            []

        ::

            sage: # needs sage.libs.pari
            sage: Q = QuadraticForm(ZZ, 2, [1,0,5], unsafe_initialization=False,
            ....:                   number_of_automorphisms=3, determinant=0)
            sage: Q.list_external_initializations()
            []
            sage: Q = QuadraticForm(ZZ, 2, [1,0,5], unsafe_initialization=True,
            ....:                   number_of_automorphisms=3, determinant=0)
            sage: Q.list_external_initializations()
            ['number_of_automorphisms', 'determinant']
        """
        return deepcopy(self._external_initialization_list)

    def __pari__(self):
        """
        Return a PARI-formatted Hessian matrix for Q.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,0,5])
            sage: Q.__pari__()                                                          # needs sage.libs.pari
            [2, 0; 0, 10]
        """
        return self.matrix().__pari__()

    def _pari_init_(self) -> str:
        """
        Return a PARI-formatted Hessian matrix for Q, as string.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,0,5])
            sage: Q._pari_init_()                                                       # needs sage.libs.pari
            'Mat([2,0;0,10])'
        """
        return self.matrix()._pari_init_()

    def _repr_(self) -> str:
        """
        Give a text representation for the quadratic form given as an upper-triangular matrix of coefficients.

        EXAMPLES::

            sage: QuadraticForm(ZZ, 2, [1,3,5])
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 1 3 ]
            [ * 5 ]
        """
        n = self.dim()
        out_str = "Quadratic form in " + str(n) + " variables over " + str(self.base_ring()) + " with coefficients: \n"
        for i in range(n):
            if i > 0:
                out_str += '\n'
            out_str += "[ "
            for j in range(n):
                if i > j:
                    out_str += "* "
                else:
                    out_str += str(self[i, j]) + " "
            out_str += "]"
        return out_str

    def _latex_(self) -> str:
        """
        Give a LaTeX representation for the quadratic form given as an upper-triangular matrix of coefficients.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [2,3,5])
            sage: Q._latex_()
            'Quadratic form in 2 variables over Integer Ring with coefficients: \\newline\\left[ \\begin{array}{cc}2 & 3 &  * & 5 & \\end{array} \\right]'
        """
        n = self.dim()
        out_str = ""
        out_str += "Quadratic form in " + str(n) + " variables over " + str(self.base_ring())
        out_str += " with coefficients: \\newline"
        out_str += "\\left[ \\begin{array}{" + n * "c" + "}"
        for i in range(n):
            for j in range(n):
                if i > j:
                    out_str += " * & "
                else:
                    out_str += str(self[i, j]) + " & "
#            if i < (n-1):
#                out_str += "\\"
        out_str += "\\end{array} \\right]"
        return out_str

    def __getitem__(self, ij):
        r"""
        Return the coefficient `a_{ij}` of `x_i\cdot x_j`.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 3, [1,2,3,4,5,6])
            sage: matrix(ZZ, 3, 3, [Q[i,j]  for i in range(3) for j in range(3)])
            [1 2 3]
            [2 4 5]
            [3 5 6]
        """
        # Unpack the list of indices
        i, j = ij
        i = int(i)
        j = int(j)

        # Ensure we're using upper-triangular coordinates
        if i > j:
            i, j = j, i

        return self.__coeffs[i * self.__n - i * (i - 1) // 2 + j - i]

    def __setitem__(self, ij, coeff):
        r"""
        Set the coefficient `a_{ij}` in front of `x_i\cdot x_j`.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 3, [1,2,3,4,5,6])
            sage: Q
            Quadratic form in 3 variables over Integer Ring with coefficients:
            [ 1 2 3 ]
            [ * 4 5 ]
            [ * * 6 ]
            sage: Q[2,1] = 17
            sage: Q
            Quadratic form in 3 variables over Integer Ring with coefficients:
            [ 1 2 3 ]
            [ * 4 17 ]
            [ * * 6 ]
        """
        # Unpack the list of indices
        i, j = ij
        i = int(i)
        j = int(j)

        # TO DO:  Verify that 0 <= i, j <= (n-1)

        # Ensure we're using upper-triangular coordinates
        if i > j:
            i, j = j, i

        # Set the entry
        try:
            self.__coeffs[i * self.__n - i * (i - 1) // 2 + j - i] = self.__base_ring(coeff)
        except Exception:
            raise RuntimeError("this coefficient cannot be coerced to an element of the base ring for the quadratic form")

    def __hash__(self) -> int:
        r"""
        TESTS::

            sage: Q1 = QuadraticForm(QQ, 2, [1,1,1])
            sage: Q2 = QuadraticForm(QQ, 2, [1,1,1])
            sage: Q3 = QuadraticForm(QuadraticField(2), 2, [1,1,1])                     # needs sage.rings.number_field
            sage: hash(Q1) == hash(Q2)
            True
            sage: hash(Q1) == hash(Q3)                                                  # needs sage.rings.number_field
            False
        """
        return hash(self.__base_ring) ^ hash(tuple(self.__coeffs))

    def __eq__(self, right) -> bool:
        """
        Determines if two quadratic forms are equal.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,4,10])
            sage: Q == Q
            True

            sage: Q1 = QuadraticForm(QQ, 2, [1,4,10])
            sage: Q == Q1
            False

            sage: Q2 = QuadraticForm(ZZ, 2, [1,4,-10])
            sage: Q == Q1
            False
            sage: Q == Q2
            False
            sage: Q1 == Q2
            False
        """
        if not isinstance(right, QuadraticForm):
            return False
        return (self.__base_ring == right.__base_ring) and (self.__coeffs == right.__coeffs)

    def __add__(self, right):
        """
        Return the direct sum of two quadratic forms.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,4,10]); Q
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 1 4 ]
            [ * 10 ]
            sage: Q2 = QuadraticForm(ZZ, 2, [1,4,-10])
            sage: Q + Q2
            Quadratic form in 4 variables over Integer Ring with coefficients:
            [ 1 4 0 0 ]
            [ * 10 0 0 ]
            [ * * 1 4 ]
            [ * * * -10 ]
        """
        if not isinstance(right, QuadraticForm):
            raise TypeError("cannot add these objects since they are not both quadratic forms")
        elif self.base_ring() != right.base_ring():
            raise TypeError("cannot add these since the quadratic forms do not have the same base rings")

        Q = QuadraticForm(self.base_ring(), self.dim() + right.dim())
        n = self.dim()
        m = right.dim()

        for i in range(n):
            for j in range(i, n):
                Q[i, j] = self[i, j]

        for i in range(m):
            for j in range(i, m):
                Q[n + i, n + j] = right[i, j]

        return Q

    def sum_by_coefficients_with(self, right):
        """
        Return the sum (on coefficients) of two quadratic forms of the same size.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,4,10]); Q
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 1 4 ]
            [ * 10 ]
            sage: Q + Q
            Quadratic form in 4 variables over Integer Ring with coefficients:
            [ 1 4 0 0 ]
            [ * 10 0 0 ]
            [ * * 1 4 ]
            [ * * * 10 ]

            sage: Q2 = QuadraticForm(ZZ, 2, [1,4,-10])
            sage: Q.sum_by_coefficients_with(Q2)
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 2 8 ]
            [ * 0 ]
        """
        if not isinstance(right, QuadraticForm):
            raise TypeError("cannot add these objects since they are not both quadratic forms")
        elif self.__n != right.__n:
            raise TypeError("cannot add these since the quadratic forms do not have the same sizes")
        elif self.__base_ring != right.__base_ring:
            raise TypeError("cannot add these since the quadratic forms do not have the same base rings")
        return QuadraticForm(self.__base_ring, self.__n, [self.__coeffs[i] + right.__coeffs[i] for i in range(len(self.__coeffs))])

    # ========================  CHANGE THIS TO A TENSOR PRODUCT?!?  Even in Characteristic 2?!?  =======================
    #    def __mul__(self, right):
    #        """
    #        Multiply (on the right) the quadratic form Q by an element of the ring that Q is defined over.
    #
    #        EXAMPLES::
    #
    #            sage: Q = QuadraticForm(ZZ, 2, [1,4,10])
    #            sage: Q*2
    #            Quadratic form in 2 variables over Integer Ring with coefficients:
    #            [ 2 8 ]
    #            [ * 20 ]
    #
    #            sage: Q+Q == Q*2
    #            True
    #        """
    #        try:
    #            c = self.base_ring()(right)
    #        except Exception:
    #            raise TypeError("the multiplier cannot be coerced into the base ring of the quadratic form")
    #        return QuadraticForm(self.base_ring(), self.dim(), [c * self.__coeffs[i]  for i in range(len(self.__coeffs))])

    def __call__(self, v):
        r"""
        Evaluate this quadratic form `Q` on a vector or matrix of elements
        coercible to the base ring of the quadratic form.

        If a vector is given then the output will be the ring element
        `Q(v)`, but if a matrix is given then the output will be the
        quadratic form `Q'` which in matrix notation is given by:

        .. MATH::

            Q' = v^t\cdot Q\cdot v.

        EXAMPLES:

        Evaluate a quadratic form at a vector::

            sage: Q = QuadraticForm(QQ, 3, range(6))
            sage: Q
            Quadratic form in 3 variables over Rational Field with coefficients:
            [ 0 1 2 ]
            [ * 3 4 ]
            [ * * 5 ]
            sage: Q([1,2,3])
            89
            sage: Q([1,0,0])
            0
            sage: Q([1,1,1])
            15

        Evaluate a quadratic form using a column matrix::

            sage: Q = QuadraticForm(QQ, 2, range(1,4))
            sage: A = Matrix(ZZ,2,2,[-1,0,0,1])
            sage: Q(A)
            Quadratic form in 2 variables over Rational Field with coefficients:
            [ 1 -2 ]
            [ * 3 ]
            sage: Q([1,0])
            1
            sage: type(Q([1,0]))
            <... 'sage.rings.rational.Rational'>
            sage: Q = QuadraticForm(QQ, 2, range(1,4))
            sage: Q(matrix(2, [1,0]))
            Quadratic form in 1 variables over Rational Field with coefficients:
            [ 1 ]

        Simple 2x2 change of variables::

            sage: Q = QuadraticForm(ZZ, 2, [1,0,1])
            sage: Q
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 1 0 ]
            [ * 1 ]
            sage: M = Matrix(ZZ, 2, 2, [1,1,0,1])
            sage: M
            [1 1]
            [0 1]
            sage: Q(M)
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 1 2 ]
            [ * 2 ]

        Some more tests::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
            sage: Q([1,2,3])
            14
            sage: v = vector([1,2,3])
            sage: Q(v)
            14
            sage: t = tuple([1,2,3])
            sage: Q(v)
            14
            sage: M = Matrix(ZZ, 3, [1,2,3])
            sage: Q(M)
            Quadratic form in 1 variables over Integer Ring with coefficients:
            [ 14 ]
        """
        # If we are passed a matrix A, return the quadratic form Q(A(x))
        # (In matrix notation: A^t * Q * A)
        n = self.dim()

        if isinstance(v, Matrix):
            # Check that v has the correct number of rows
            if v.nrows() != n:
                raise TypeError(f"the matrix must have {n} rows")

            # Create the new quadratic form
            m = v.ncols()
            Q2 = QuadraticForm(self.base_ring(), m)
            return QFEvaluateMatrix(self, v, Q2)

        if isinstance(v, (Vector, list, tuple)):
            # Check the vector/tuple/list has the correct length
            if len(v) != n:
                raise TypeError(f"your vector needs to have length {n}")

            # TO DO:  Check that the elements can be coerced into the base ring of Q -- on first elt.
            if len(v) > 0:
                try:
                    self.base_ring()(v[0])
                except Exception:
                    raise TypeError("your vector is not coercible to the base ring of the quadratic form")

            # Attempt to evaluate Q[v]
            return QFEvaluateVector(self, v)

        raise TypeError

    # ===============================================

    def _is_even_symmetric_matrix_(self, A, R=None):
        """
        Test if a matrix is symmetric, defined over `R`, and has even diagonal in `R`.

        INPUT:

        - ``A`` -- matrix

        - ``R`` -- ring

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [2,3,5])
            sage: A = Q.matrix()
            sage: A
            [ 4  3]
            [ 3 10]
            sage: Q._is_even_symmetric_matrix_(A)
            True
            sage: A[0,0] = 1
            sage: Q._is_even_symmetric_matrix_(A)
            False
        """
        if not isinstance(A, Matrix):
            raise TypeError("A is not a matrix.")

        ring_coerce_test = True
        if R is None:            # This allows us to omit the ring from the variables, and take it from the matrix
            R = A.base_ring()
            ring_coerce_test = False

        if R not in Rings():
            raise TypeError("R is not a ring.")

        if not (A.is_square() and A.is_symmetric()):
            return False

        # Test that all entries coerce to R
        n = A.nrows()
        if not (A.base_ring() == R or ring_coerce_test):
            try:
                for i in range(n):
                    for j in range(i, n):
                        R(A[i, j])
            except (TypeError, ValueError):
                return False

        # Test that the diagonal is even (if 1/2 isn't in R)
        if not R(2).is_unit():
            for i in range(n):
                if not is_even(R(A[i, i])):
                    return False

        return True

    # =====================================================================

    def matrix(self):
        r"""
        Return the Hessian matrix `A` for which `Q(X) = (1/2) X^t\cdot A\cdot X`.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 3, range(6))
            sage: Q.matrix()
            [ 0  1  2]
            [ 1  6  4]
            [ 2  4 10]
        """
        return self.Hessian_matrix()

    def Hessian_matrix(self):
        r"""
        Return the Hessian matrix `A` for which `Q(X) = (1/2) X^t\cdot A\cdot X`.

        EXAMPLES::

            sage: Q = QuadraticForm(QQ, 2, range(1,4)); Q
            Quadratic form in 2 variables over Rational Field with coefficients:
            [ 1 2 ]
            [ * 3 ]
            sage: Q.Hessian_matrix()
            [2 2]
            [2 6]
            sage: Q.matrix().base_ring()
            Rational Field
        """
        mat_entries = []
        for i in range(self.dim()):
            for j in range(self.dim()):
                if i == j:
                    mat_entries += [2 * self[i, j]]
                else:
                    mat_entries += [self[i, j]]

        return matrix(self.base_ring(), self.dim(), self.dim(), mat_entries)

    def Gram_matrix_rational(self):
        r"""
        Return a (symmetric) Gram matrix `A` for the quadratic form `Q`,
        meaning that

        .. MATH::

            Q(x) = x^t\cdot A\cdot x,

        defined over the fraction field of the base ring.

        EXAMPLES::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
            sage: A = Q.Gram_matrix_rational(); A
            [1 0 0 0]
            [0 3 0 0]
            [0 0 5 0]
            [0 0 0 7]
            sage: A.base_ring()
            Rational Field
        """
        return (ZZ.one() / ZZ(2)) * self.matrix()

    def Gram_matrix(self):
        r"""
        Return a (symmetric) Gram matrix `A` for the quadratic form `Q`,
        meaning that

        .. MATH::

            Q(x) = x^t\cdot A\cdot x,

        defined over the base ring of `Q`.  If this is not possible,
        then a :exc:`TypeError` is raised.

        EXAMPLES::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
            sage: A = Q.Gram_matrix(); A
            [1 0 0 0]
            [0 3 0 0]
            [0 0 5 0]
            [0 0 0 7]
            sage: A.base_ring()
            Integer Ring
        """
        A = (ZZ.one() / ZZ(2)) * self.matrix()
        n = self.dim()

        # Test to see if it has an integral Gram matrix
        Int_flag = True
        for i in range(n):
            for j in range(i, n):
                Int_flag &= A[i, j] in self.base_ring()

        # Return the Gram matrix, or an error
        if Int_flag:
            return MatrixSpace(self.base_ring(), n, n)(A)
        raise TypeError("this form does not have an integral Gram matrix")

    def has_integral_Gram_matrix(self) -> bool:
        r"""
        Return whether the quadratic form has an integral Gram matrix (with respect to its base ring).

        A warning is issued if the form is defined over a field,
        since in that case the return is trivially true.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [7,8,9])
            sage: Q.has_integral_Gram_matrix()
            True

        ::

            sage: Q = QuadraticForm(ZZ, 2, [4,5,6])
            sage: Q.has_integral_Gram_matrix()
            False
        """
        # Warning over fields
        if self.base_ring() in Fields():
            warn("Warning -- A quadratic form over a field always has integral Gram matrix.  Do you really want to do this?!?")

        # Determine integrality of the Gram matrix
        try:
            self.Gram_matrix()
        except TypeError:
            return False
        else:
            return True

    def gcd(self):
        """
        Return the greatest common divisor of the coefficients of the
        quadratic form (as a polynomial).

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 4, range(1, 21, 2))
            sage: Q.gcd()
            1

            sage: Q = QuadraticForm(ZZ, 4, range(0, 20, 2))
            sage: Q.gcd()
            2
        """
        if self.base_ring() != ZZ:
            raise TypeError("the given quadratic form must be defined over ZZ")
        return GCD(self.coefficients())

    def polynomial(self, names='x'):
        r"""
        Return the quadratic form as a polynomial in `n` variables.

        INPUT:

        - ``self`` -- a quadratic form over a commutative ring

        - ``names`` -- specification of the names of the variables; see :func:`PolynomialRing`

        OUTPUT: the polynomial form of the quadratic form

        EXAMPLES::

            sage: Q = DiagonalQuadraticForm(QQ,[1, 3, 5, 7])
            sage: P = Q.polynomial(); P
            x0^2 + 3*x1^2 + 5*x2^2 + 7*x3^2

        ::

            sage: # needs sage.rings.number_field
            sage: x = polygen(ZZ, 'x')
            sage: F.<a> = NumberField(x^2 - 5)
            sage: Z = F.ring_of_integers()
            sage: Q = QuadraticForm(Z, 3, [2*a, 3*a, 0, 1 - a, 0, 2*a + 4])
            sage: P = Q.polynomial(names='y'); P
            2*a*y0^2 + 3*a*y0*y1 + (-a + 1)*y1^2 + (2*a + 4)*y2^2
            sage: Q = QuadraticForm(F, 4,
            ....:                   [a, 3*a, 0, 1 - a, a - 3, 0, 2*a + 4, 4 + a, 0, 1])
            sage: Q.polynomial(names='z')
            a*z0^2 + (3*a)*z0*z1 + (a - 3)*z1^2 + (a + 4)*z2^2
            + (-a + 1)*z0*z3 + (2*a + 4)*z1*z3 + z3^2
            sage: B.<i,j,k> = QuaternionAlgebra(F,-1,-1)
            sage: Q = QuadraticForm(B, 3, [2*a, 3*a, i, 1 - a, 0, 2*a + 4])
            sage: Q.polynomial()
            Traceback (most recent call last):
            ...
            ValueError: Can only create polynomial rings over commutative rings
        """
        B = self.base_ring()
        if B not in Rings().Commutative():
            raise ValueError('Can only create polynomial rings over commutative rings')
        n = self.dim()
        M = matrix(B, n)
        for i in range(n):
            for j in range(i, n):
                M[i, j] = self[i, j]
        R = PolynomialRing(self.base_ring(), names, n)
        V = vector(R.gens())
        return (V * M).dot_product(V)

    @staticmethod
    def from_polynomial(poly):
        r"""
        Construct a :class:`QuadraticForm` from a multivariate
        polynomial. Inverse of :meth:`polynomial`.

        EXAMPLES::

            sage: R.<x,y,z> = ZZ[]
            sage: f = 5*x^2 - x*z - 3*y*z - 2*y^2 + 9*z^2
            sage: Q = QuadraticForm.from_polynomial(f); Q
            Quadratic form in 3 variables over Integer Ring with coefficients:
            [ 5 0 -1 ]
            [ * -2 -3 ]
            [ * * 9 ]
            sage: Q.polynomial()
            5*x0^2 - 2*x1^2 - x0*x2 - 3*x1*x2 + 9*x2^2
            sage: Q.polynomial()(R.gens()) == f
            True

        The method fails if the given polynomial is not a quadratic form::

            sage: QuadraticForm.from_polynomial(x^3 + x*z + 5*y^2)
            Traceback (most recent call last):
            ...
            ValueError: polynomial has monomials of degree != 2
        """
        R = poly.parent()
        from sage.rings.polynomial.multi_polynomial_ring_base import (
            MPolynomialRing_base,
        )
        if not isinstance(R, MPolynomialRing_base):
            raise TypeError(f'not a multivariate polynomial ring: {R}')
        if not all(mon.degree() == 2 for mon in poly.monomials()):
            raise ValueError('polynomial has monomials of degree != 2')
        base = R.base_ring()
        vs = R.gens()
        coeffs = [poly.monomial_coefficient(v * w)
                  for i, v in enumerate(vs) for w in vs[i:]]
        return QuadraticForm(base, len(vs), coeffs)

    def is_primitive(self) -> bool:
        """
        Determine if the given integer-valued form is primitive.

        This means not an integer (`> 1`) multiple of another integer-valued
        quadratic form.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [2,3,4])
            sage: Q.is_primitive()
            True
            sage: Q = QuadraticForm(ZZ, 2, [2,4,8])
            sage: Q.is_primitive()
            False
        """
        return self.gcd() == 1

    def primitive(self):
        r"""
        Return a primitive version of an integer-valued quadratic form, defined over `\ZZ`.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [2,3,4])
            sage: Q.primitive()
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 2 3 ]
            [ * 4 ]
            sage: Q = QuadraticForm(ZZ, 2, [2,4,8])
            sage: Q.primitive()
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 1 2 ]
            [ * 4 ]
        """
        if self.base_ring() != ZZ:
            raise TypeError("the given quadratic form must be defined over ZZ")
        g = self.gcd()
        return QuadraticForm(ZZ, self.dim(),
                             [x // g for x in self.coefficients()])

    def adjoint_primitive(self):
        """
        Return the primitive adjoint of the quadratic form, which is
        the smallest discriminant integer-valued quadratic form whose
        matrix is a scalar multiple of the inverse of the matrix of
        the given quadratic form.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,2,3])
            sage: Q.adjoint_primitive()
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 3 -2 ]
            [ *  1 ]
        """
        return QuadraticForm(self.Hessian_matrix().adjoint_classical()).primitive()

    def dim(self):
        """
        Return the number of variables of the quadratic form.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,2,3])
            sage: Q.dim()
            2
            sage: parent(Q.dim())
            Integer Ring
            sage: Q = QuadraticForm(Q.matrix())
            sage: Q.dim()
            2
            sage: parent(Q.dim())
            Integer Ring
        """
        return self.__n

    def base_ring(self):
        """
        Return the ring over which the quadratic form is defined.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,2,3])
            sage: Q.base_ring()
            Integer Ring
        """
        return self.__base_ring

    def coefficients(self):
        r"""
        Return the matrix of upper triangular coefficients,
        by reading across the rows from the main diagonal.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,2,3])
            sage: Q.coefficients()
            [1, 2, 3]
        """
        return self.__coeffs

    def det(self):
        r"""
        Return the determinant of the Gram matrix of `2\cdot Q`, or
        equivalently the determinant of the Hessian matrix of `Q`.

        .. NOTE::

            This is always defined over the same ring as the quadratic form.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,2,3])
            sage: Q.det()
            8
        """
        try:
            return self.__det
        except AttributeError:
            # Compute the determinant
            if self.dim() == 0:
                new_det = self.base_ring()(1)
            else:
                new_det = self.matrix().det()

            # Cache and return the determinant
            self.__det = new_det
            return new_det

    def Gram_det(self):
        r"""
        Return the determinant of the Gram matrix of `Q`.

        .. NOTE::

            This is defined over the fraction field of the ring of
            the quadratic form, but is often not defined over the same
            ring as the quadratic form.

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, [1,2,3])
            sage: Q.Gram_det()
            2
        """
        return self.det() / ZZ(2**self.dim())

    def change_ring(self, R):
        """
        Alters the quadratic form to have all coefficients
        defined over the new base ring `R`.  Here `R` must be
        coercible to from the current base ring.

        .. NOTE::

            This is preferable to performing an explicit
            coercion through the :meth:`base_ring` method, which does
            not affect the individual coefficients.  This is
            particularly useful for performing fast modular
            arithmetic evaluations.

        INPUT:

        - ``R`` -- a ring

        OUTPUT: quadratic form

        EXAMPLES::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,1]); Q
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 1 0 ]
            [ * 1 ]

            sage: Q1 = Q.change_ring(IntegerModRing(5)); Q1
            Quadratic form in 2 variables over Ring of integers modulo 5 with coefficients:
            [ 1 0 ]
            [ * 1 ]

            sage: Q1([35,11])
            1
        """
        # Check that a canonical coercion is possible
        if R not in Rings():
            raise TypeError("R is not a ring")
        if not R.has_coerce_map_from(self.base_ring()):
            raise TypeError(f"there is no canonical coercion from {self.base_ring()} to R")
        # Return the coerced form
        return QuadraticForm(R, self.dim(), [R(x) for x in self.coefficients()])

    def level(self):
        r"""
        Determine the level of the quadratic form over a PID.

        This is a generator for the smallest ideal `N` of `R` such
        that `N\cdot (` the matrix of `2*Q` `)^{(-1)}` is in `R` with
        diagonal in `2R`.

        Over `\ZZ` this returns a nonnegative number.

        (Caveat: This always returns the unit ideal when working over a field!)

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, range(1,4))
            sage: Q.level()
            8

            sage: Q1 = QuadraticForm(QQ, 2, range(1,4))
            sage: Q1.level()      # random
            UserWarning: Warning -- The level of a quadratic form over a field is always 1.
            Do you really want to do this?!?
            1

            sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
            sage: Q.level()
            420
        """
        # Try to return the cached level
        try:
            return self.__level
        except AttributeError:

            # Check that the base ring is a PID
            if self.base_ring() not in PrincipalIdealDomains():
                raise TypeError("the level (as a number) is only defined over a Principal Ideal Domain ; try using level_ideal()")

            # Warn the user if the form is defined over a field!
            if self.base_ring() in Fields():
                warn("Warning -- The level of a quadratic form over a field is always 1.  Do you really want to do this?!?")
                # raise RuntimeError("Warning -- The level of a quadratic form over a field is always 1.  Do you really want to do this?!?")

            # Check invertibility and find the inverse
            try:
                mat_inv = self.matrix()**(-1)
            except ZeroDivisionError:
                raise TypeError("the quadratic form is degenerate")

            # Compute the level
            inv_denoms = []
            for i in range(self.dim()):
                for j in range(i, self.dim()):
                    if i == j:
                        inv_denoms += [denominator(mat_inv[i, j] / 2)]
                    else:
                        inv_denoms += [denominator(mat_inv[i, j])]
            lvl = LCM(inv_denoms)
            lvl = Ideal(self.base_ring()(lvl)).gen()
            ##############################################################
            # To do this properly, the level should be the inverse of the
            # fractional ideal (over R) generated by the entries whose
            # denominators we take above. =)
            ##############################################################

            # Normalize the result over ZZ
            if self.base_ring() == ZZ:
                lvl = abs(lvl)

            # Cache and return the level
            self.__level = lvl
            return lvl

    def level_ideal(self):
        r"""
        Determine the level of the quadratic form (over `R`).

        This is the smallest ideal `N` of `R` such that `N \cdot (`
        the matrix of `2Q` `)^{(-1)}` is in `R` with diagonal in `2R`.

        (Caveat: This always returns the principal ideal when working
        over a field!)

        .. WARNING::

            This only works over a PID ring of integers for now!
            (Waiting for Sage fractional ideal support.)

        EXAMPLES::

            sage: Q = QuadraticForm(ZZ, 2, range(1,4))
            sage: Q.level_ideal()
            Principal ideal (8) of Integer Ring

            sage: Q1 = QuadraticForm(QQ, 2, range(1,4))
            sage: Q1.level_ideal()
            Principal ideal (1) of Rational Field

            sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
            sage: Q.level_ideal()
            Principal ideal (420) of Integer Ring
        """
        ##############################################################
        # To do this properly, the level should be the inverse of the
        # fractional ideal (over R) generated by the entries whose
        # denominators we take above.
        ##############################################################
        return Ideal(self.base_ring()(self.level()))

    def bilinear_map(self, v, w):
        r"""
        Return the value of the associated bilinear map on two vectors.

        Given a quadratic form `Q` over some base ring `R` with
        characteristic not equal to 2, this gives the image of two
        vectors with coefficients in `R` under the associated bilinear
        map `B`, given by the relation `2 B(v,w) = Q(v) + Q(w) - Q(v+w)`.

        INPUT:

        - ``v``, ``w`` -- two vectors

        OUTPUT: an element of the base ring `R`

        EXAMPLES:

        First, an example over `\ZZ`::

            sage: Q = QuadraticForm(ZZ, 3, [1,4,0,1,4,1])
            sage: v = vector(ZZ, (1,2,0))
            sage: w = vector(ZZ, (0,1,1))
            sage: Q.bilinear_map(v, w)
            8

        This also works over `\QQ`::

            sage: Q = QuadraticForm(QQ, 2, [1/2,2,1])
            sage: v = vector(QQ, (1,1))
            sage: w = vector(QQ, (1/2,2))
            sage: Q.bilinear_map(v, w)
            19/4

        The vectors must have the correct length::

            sage: Q = DiagonalQuadraticForm(ZZ, [1,7,7])
            sage: v = vector((1,2))
            sage: w = vector((1,1,1))
            sage: Q.bilinear_map(v, w)
            Traceback (most recent call last):
            ...
            TypeError: vectors must have length 3

        This does not work if the characteristic is 2::

            sage: # needs sage.rings.finite_rings
            sage: Q = DiagonalQuadraticForm(GF(2), [1,1,1])
            sage: v = vector((1,1,1))
            sage: w = vector((1,1,1))
            sage: Q.bilinear_map(v, w)
            Traceback (most recent call last):
            ...
            TypeError: not defined for rings of characteristic 2
        """
        if len(v) != self.dim() or len(w) != self.dim():
            raise TypeError("vectors must have length " + str(self.dim()))
        if self.base_ring().characteristic() == 2:
            raise TypeError("not defined for rings of characteristic 2")
        return (self(v + w) - self(v) - self(w)) / 2


def DiagonalQuadraticForm(R, diag) -> QuadraticForm:
    """
    Return a quadratic form over `R` which is a sum of squares.

    INPUT:

    - ``R`` -- ring
    - ``diag`` -- list/tuple of elements coercible to `R`

    OUTPUT: quadratic form

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7]); Q
        Quadratic form in 4 variables over Integer Ring with coefficients:
        [ 1 0 0 0 ]
        [ * 3 0 0 ]
        [ * * 5 0 ]
        [ * * * 7 ]
    """
    Q = QuadraticForm(R, len(diag))
    for i in range(len(diag)):
        Q[i, i] = diag[i]
    return Q



r"""
AUTHORS:

- Jonas Jermann (2013): initial version
"""
# ****************************************************************************
#       Copyright (C) 2013-2014 Jonas Jermann <jjermann2@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.modular.modform_hecketriangle.hecke_triangle_groups import HeckeTriangleGroup

from sage.modular.modform_hecketriangle.series_constructor import MFSeriesConstructor

from sage.modular.modform_hecketriangle.graded_ring import (QuasiMeromorphicModularFormsRing,
                                                            QuasiWeakModularFormsRing, QuasiModularFormsRing,
                                                            QuasiCuspFormsRing, MeromorphicModularFormsRing,
                                                            WeakModularFormsRing,
                                                            ModularFormsRing, CuspFormsRing)

from sage.modular.modform_hecketriangle.space import (QuasiMeromorphicModularForms, QuasiWeakModularForms,
                                                      QuasiModularForms, QuasiCuspForms,
                                                      MeromorphicModularForms, WeakModularForms, ModularForms,
                                                      CuspForms, ZeroForm)

from sage.modular.modform_hecketriangle.subspace import ModularFormsSubSpace



r"""
The matrix monoid `\Sigma_0(N)`.

This stands for a monoid of matrices over `\ZZ`, `\QQ`, `\ZZ_p`, or `\QQ_p`,
depending on an integer `N \ge 1`. This class exists in order to act on `p`-adic
distribution spaces.

Over `\QQ` or `\ZZ`, it is the monoid of matrices `2\times2` matrices
`\begin{pmatrix} a & b \\ c & d \end{pmatrix}`
such that

- `ad - bc \ne 0`,
- `a` is integral and invertible at the primes dividing `N`,
- `c` has valuation at least `v_p(N)` for each `p` dividing `N` (but may be
  non-integral at other places).

The value `N=1` is allowed, in which case the second and third conditions are vacuous.

EXAMPLES::

    sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
    sage: S1 = Sigma0(1); S3 = Sigma0(3)
    sage: S1([3, 0, 0, 1])
    [3 0]
    [0 1]
    sage: S3([3, 0, 0, 1]) # boom
    Traceback (most recent call last):
    ...
    TypeError: 3 is not a unit at 3
    sage: S3([5,0,0,1])
    [5 0]
    [0 1]
    sage: S3([1, 0, 0, 3])
    [1 0]
    [0 3]
    sage: matrix(ZZ, 2, [1,0,0,1]) in S1
    True

AUTHORS:

    - David Pollack (2012): initial version
"""

# Warning to developers: when working with Sigma0 elements it is generally a
# good idea to avoid using the entries of x.matrix() directly; rather, use the
# "adjuster" mechanism. The purpose of this is to allow us to seamlessly change
# conventions for matrix actions (since there are several in use in the
# literature and no natural "best" choice).
from sage.matrix.matrix_space import MatrixSpace
from sage.misc.abstract_method import abstract_method
from sage.structure.factory import UniqueFactory
from sage.structure.element import MonoidElement
from sage.structure.richcmp import richcmp
from sage.categories.monoids import Monoids
from sage.categories.morphism import Morphism
from sage.structure.parent import Parent
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.structure.unique_representation import UniqueRepresentation


class Sigma0ActionAdjuster(UniqueRepresentation):

    @abstract_method
    def __call__(self, x):
        r"""
        Given a :class:`Sigma0element` ``x``, return four integers.

        This is used to allow for other conventions for the action of Sigma0
        on the space of distributions.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import _default_adjuster
            sage: A = _default_adjuster()
            sage: A(matrix(ZZ, 2, [3,4,5,6])) # indirect doctest
            (3, 4, 5, 6)
        """
        pass


class _default_adjuster(Sigma0ActionAdjuster):
    """
    A callable object that does nothing to a matrix, returning its entries
    in the natural, by-row, order.

    INPUT:

    - ``g`` -- a `2 \times 2` matrix

    OUTPUT: a 4-tuple consisting of the entries of the matrix

    EXAMPLES::

        sage: A = sage.modular.pollack_stevens.sigma0._default_adjuster(); A
        <sage.modular.pollack_stevens.sigma0._default_adjuster object at 0x...>
        sage: TestSuite(A).run()
    """
    def __call__(self, g):
        """
        EXAMPLES::

            sage: T = sage.modular.pollack_stevens.sigma0._default_adjuster()
            sage: T(matrix(ZZ,2,[1..4])) # indirect doctest
            (1, 2, 3, 4)
        """
        return tuple(g.list())


class Sigma0_factory(UniqueFactory):
    r"""
    Create the monoid of non-singular matrices, upper triangular mod `N`.

    INPUT:

    - ``N`` -- integer; the level (should be strictly positive)
    - ``base_ring`` (commutative ring, default `\ZZ`) -- the base
      ring (normally `\ZZ` or a `p`-adic ring)
    - ``adjuster`` -- ``None``, or a callable which takes a `2 \times 2` matrix
      and returns a 4-tuple of integers. This is supplied in order to support
      differing conventions for the action of `2 \times 2` matrices on
      distributions.

    EXAMPLES::

        sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
        sage: Sigma0(3)
        Monoid Sigma0(3) with coefficients in Integer Ring
    """

    def create_key(self, N, base_ring=ZZ, adjuster=None):
        r"""
        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: Sigma0.create_key(3)
            (3, Integer Ring, <sage.modular.pollack_stevens.sigma0._default_adjuster object at 0x...>)
            sage: TestSuite(Sigma0).run()
        """
        N = ZZ(N)
        if N <= 0:
            raise ValueError("Modulus should be > 0")
        if adjuster is None:
            adjuster = _default_adjuster()

        if base_ring not in (QQ, ZZ):
            try:
                if not N.is_power_of(base_ring.prime()):
                    raise ValueError("Modulus must equal base ring prime")
            except AttributeError:
                raise ValueError("Base ring must be QQ, ZZ or a p-adic field")
        return (N, base_ring, adjuster)

    def create_object(self, version, key):
        r"""
        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: Sigma0(3) # indirect doctest
            Monoid Sigma0(3) with coefficients in Integer Ring
        """
        return Sigma0_class(*key)


Sigma0 = Sigma0_factory('sage.modular.pollack_stevens.sigma0.Sigma0')


class Sigma0Element(MonoidElement):
    r"""
    An element of the monoid Sigma0. This is a wrapper around a `2 \times 2` matrix.

    EXAMPLES::

        sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
        sage: S = Sigma0(7)
        sage: g = S([2,3,7,1])
        sage: g.det()
        -19
        sage: h = S([1,2,0,1])
        sage: g * h
        [ 2  7]
        [ 7 15]
        sage: g.inverse()
        Traceback (most recent call last):
        ...
        TypeError: no conversion of this rational to integer
        sage: h.inverse()
        [ 1 -2]
        [ 0  1]
    """
    def __init__(self, parent, mat):
        r"""
        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: s = Sigma0(3)([1,4,3,3]) # indirect doctest
            sage: TestSuite(s).run()
        """
        self._mat = mat
        MonoidElement.__init__(self, parent)

    def __hash__(self):
        r"""
        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: s = Sigma0(3)([1,4,3,3])
            sage: hash32 = 619049499
            sage: hash64 = 8095169151987216923
            sage: hash(s) in [hash32, hash64]  # indirect doctest
            True
        """
        return hash(self.matrix())

    def det(self):
        r"""
        Return the determinant of this matrix, which is (by assumption) nonzero.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: s = Sigma0(3)([1,4,3,3])
            sage: s.det()
            -9
        """
        return self.matrix().det()

    def _mul_(self, other):
        r"""
        Return the product of two Sigma0 elements.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: s = Sigma0(3)([1,4,3,3])
            sage: t = Sigma0(15)([4,0,0,1])
            sage: u = s*t; u # indirect doctest
            [ 4  4]
            [12  3]
            sage: type(u)
            <class 'sage.modular.pollack_stevens.sigma0.Sigma0_class_with_category.element_class'>
            sage: u.parent()
            Monoid Sigma0(3) with coefficients in Integer Ring
        """
        return self.parent()(self._mat * other._mat, check=False)

    def _richcmp_(self, other, op):
        r"""
        Compare two elements (of a common Sigma0 object).

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: s = Sigma0(3)([1,4,3,3])
            sage: t = Sigma0(3)([4,0,0,1])
            sage: s == t
            False
            sage: s == Sigma0(1)([1,4,3,3])
            True

        This uses the coercion model to find a common parent, with occasionally surprising results::

            sage: t == Sigma0(5)([4, 0, 0, 1])
            False
        """
        return richcmp(self._mat, other._mat, op)

    def _repr_(self):
        r"""
        String representation of ``self``.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: s = Sigma0(3)([1,4,3,3])
            sage: s._repr_()
            '[1 4]\n[3 3]'
        """
        return self.matrix().__repr__()

    def matrix(self):
        r"""
        Return ``self`` as a matrix (forgetting the additional data that it is
        in ``Sigma0(N)``).

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: s = Sigma0(3)([1,4,3,3])
            sage: sm = s.matrix()
            sage: type(s)
            <class 'sage.modular.pollack_stevens.sigma0.Sigma0_class_with_category.element_class'>
            sage: type(sm)
            <class 'sage.matrix.matrix_integer_dense.Matrix_integer_dense'>
            sage: s == sm
            True
        """
        return self._mat

    def __invert__(self):
        r"""
        Return the inverse of ``self``.

        This will raise an error if the result is not in the monoid.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: s = Sigma0(3)([1,4,3,13])
            sage: s.inverse()    # indirect doctest
            [13 -4]
            [-3  1]
            sage: Sigma0(3)([1, 0, 0, 3]).inverse()
            Traceback (most recent call last):
            ...
            TypeError: no conversion of this rational to integer

        .. TODO::

            In an ideal world this would silently extend scalars to `\QQ` if
            the inverse has non-integer entries but is still in `\Sigma_0(N)`
            locally at `N`. But we do not use such functionality, anyway.
        """
        return self.parent()(~self._mat)


class _Sigma0Embedding(Morphism):
    r"""
    A Morphism object giving the natural inclusion of `\Sigma_0` into the
    appropriate matrix space. This snippet of code is fed to the coercion
    framework so that "x * y" will work if ``x`` is a matrix and ``y`` is a `\Sigma_0`
    element (returning a matrix, *not* a Sigma0 element).
    """
    def __init__(self, domain):
        r"""
        TESTS::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0, _Sigma0Embedding
            sage: x = _Sigma0Embedding(Sigma0(3))
            sage: TestSuite(x).run(skip=['_test_category'])

        # TODO: The category test breaks because _Sigma0Embedding is not an instance of
        # the element class of its parent (a homset in the category of
        # monoids). I have no idea how to fix this.
        """
        Morphism.__init__(self, domain.Hom(domain._matrix_space, category=Monoids()))

    def _call_(self, x):
        r"""
        Return a matrix.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0, _Sigma0Embedding
            sage: S = Sigma0(3)
            sage: x = _Sigma0Embedding(S)
            sage: x(S([1,0,0,3])).parent() # indirect doctest
            Full MatrixSpace of 2 by 2 dense matrices over Integer Ring
        """
        return x.matrix()

    def _richcmp_(self, other, op):
        r"""
        Required for pickling.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0, _Sigma0Embedding
            sage: S = Sigma0(3)
            sage: x = _Sigma0Embedding(S)
            sage: x == loads(dumps(x))
            True
        """
        return richcmp(self.domain(), other.domain(), op)


class Sigma0_class(Parent):
    r"""
    The class representing the monoid `\Sigma_0(N)`.

    EXAMPLES::

        sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
        sage: S = Sigma0(5); S
        Monoid Sigma0(5) with coefficients in Integer Ring
        sage: S([1,2,1,1])
        Traceback (most recent call last):
        ...
        TypeError: level 5^1 does not divide 1
        sage: S([1,2,5,1])
        [1 2]
        [5 1]
    """
    Element = Sigma0Element

    def __init__(self, N, base_ring, adjuster):
        r"""
        Standard init function. For args documentation see the factory
        function.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: S = Sigma0(3) # indirect doctest
            sage: TestSuite(S).run()
        """
        self._N = N
        self._primes = list(N.factor())
        self._base_ring = base_ring
        self._adjuster = adjuster
        self._matrix_space = MatrixSpace(base_ring, 2)
        Parent.__init__(self, category=Monoids())
        self.register_embedding(_Sigma0Embedding(self))

    def _an_element_(self):
        r"""
        Return an element of ``self``. This is implemented in a rather dumb way.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: S = Sigma0(3)
            sage: S.an_element() # indirect doctest
            [1 0]
            [0 1]
        """
        return self([1, 0, 0, 1])

    def level(self):
        r"""
        If this monoid is `\Sigma_0(N)`, return `N`.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: S = Sigma0(3)
            sage: S.level()
            3
        """
        return self._N

    def base_ring(self):
        r"""
        Return the base ring.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: S = Sigma0(3)
            sage: S.base_ring()
            Integer Ring
        """
        return self._base_ring

    def _coerce_map_from_(self, other):
        r"""
        Find out whether ``other`` coerces into ``self``.

        The *only* thing that coerces canonically into `\Sigma_0` is another
        `\Sigma_0`. It is *very bad* if integers are allowed to coerce in, as
        this leads to a noncommutative coercion diagram whenever we let
        `\Sigma_0` act on anything..

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: Sigma0(1, QQ).has_coerce_map_from(Sigma0(3, ZZ)) # indirect doctest
            True
            sage: Sigma0(1, ZZ).has_coerce_map_from(ZZ)
            False

        (If something changes that causes the last doctest above to return
        True, then the entire purpose of this class is violated, and all sorts
        of nasty things will go wrong with scalar multiplication of
        distributions. Do not let this happen!)
        """
        return (isinstance(other, Sigma0_class)
                and self.level().divides(other.level())
                and self.base_ring().has_coerce_map_from(other.base_ring()))

    def _element_constructor_(self, x, check=True):
        r"""
        Construct an element of ``self`` from x.

        INPUT:

        - ``x`` -- something that one can make into a matrix over the
          appropriate base ring
        - ``check`` -- boolean (default: ``True``); if ``True``, then check
          that this matrix actually satisfies the conditions

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: S = Sigma0(3)
            sage: S([1,0,0,3]) # indirect doctest
            [1 0]
            [0 3]
            sage: S([3,0,0,1]) # boom
            Traceback (most recent call last):
            ...
            TypeError: 3 is not a unit at 3
            sage: S(Sigma0(1)([3,0,0,1]), check=False) # don't do this
            [3 0]
            [0 1]
        """
        if isinstance(x, Sigma0Element):
            x = x.matrix()
        if check:
            x = self._matrix_space(x)
            a, b, c, d = self._adjuster(x)
            for (p, e) in self._primes:
                if c.valuation(p) < e:
                    raise TypeError("level %s^%s does not divide %s" % (p, e, c))
                if a.valuation(p) != 0:
                    raise TypeError("%s is not a unit at %s" % (a, p))
            if x.det() == 0:
                raise TypeError("matrix must be nonsingular")
        x.set_immutable()
        return self.element_class(self, x)

    def _repr_(self):
        r"""
        String representation of ``self``.

        EXAMPLES::

            sage: from sage.modular.pollack_stevens.sigma0 import Sigma0
            sage: S = Sigma0(3)
            sage: S._repr_()
            'Monoid Sigma0(3) with coefficients in Integer Ring'
        """
        return 'Monoid Sigma0(%s) with coefficients in %s' % (self.level(),
                                                              self.base_ring())



"""
Line plots
"""

# *****************************************************************************
#       Copyright (C) 2006 Alex Clemesha <clemesha@gmail.com>,
#                          William Stein <wstein@gmail.com>,
#                     2008 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
# *****************************************************************************
from sage.plot.primitive import GraphicPrimitive_xydata
from sage.misc.decorators import options, rename_keyword
from sage.plot.colors import to_mpl_color


class Line(GraphicPrimitive_xydata):
    """
    Primitive class that initializes the line graphics type.

    EXAMPLES::

        sage: from sage.plot.line import Line
        sage: Line([1,2,7], [1,5,-1], {})
        Line defined by 3 points
    """
    def __init__(self, xdata, ydata, options):
        """
        Initialize a line graphics primitive.

        EXAMPLES::

            sage: from sage.plot.line import Line
            sage: Line([-1,2], [17,4], {'thickness':2})
            Line defined by 2 points
        """
        valid_options = self._allowed_options()
        for opt in options:
            if opt not in valid_options:
                raise RuntimeError("error in line(): option '%s' not valid" % opt)
        self.xdata = xdata
        self.ydata = ydata
        GraphicPrimitive_xydata.__init__(self, options)

    def _allowed_options(self):
        """
        Displayed the list of allowed line options.

        EXAMPLES::

            sage: from sage.plot.line import Line
            sage: list(sorted(Line([-1,2], [17,4], {})._allowed_options().items()))
            [('alpha', 'How transparent the line is.'),
             ('hue', 'The color given as a hue.'),
             ('legend_color', 'The color of the legend text.'),
             ('legend_label', 'The label for this item in the legend.'),
             ('linestyle',
              "The style of the line, which is one of '--' (dashed), '-.' (dash dot), '-' (solid), 'steps', ':' (dotted)."),
             ('marker', 'the marker symbol (see documentation for line2d for details)'),
             ('markeredgecolor', 'the color of the marker edge'),
             ('markeredgewidth', 'the size of the marker edge in points'),
             ('markerfacecolor', 'the color of the marker face'),
             ('markersize', 'the size of the marker in points'),
             ('rgbcolor', 'The color as an RGB tuple.'),
             ('thickness', 'How thick the line is.'),
             ('zorder', 'The layer level in which to draw')]
        """
        return {'alpha': 'How transparent the line is.',
                'legend_color': 'The color of the legend text.',
                'legend_label': 'The label for this item in the legend.',
                'thickness': 'How thick the line is.',
                'rgbcolor': 'The color as an RGB tuple.',
                'hue': 'The color given as a hue.',
                'linestyle': "The style of the line, which is one of '--' (dashed), '-.' (dash dot), '-' (solid), 'steps', ':' (dotted).",
                'marker': "the marker symbol (see documentation for line2d for details)",
                'markersize': 'the size of the marker in points',
                'markeredgecolor': 'the color of the marker edge',
                'markeredgewidth': 'the size of the marker edge in points',
                'markerfacecolor': 'the color of the marker face',
                'zorder': 'The layer level in which to draw'
                }

    def _plot3d_options(self, options=None):
        """
        Translate 2D plot options into 3D plot options.

        EXAMPLES::

            sage: L = line([(1,1), (1,2), (2,2), (2,1)], alpha=.5, thickness=10, zorder=2)
            sage: l=L[0]; l
            Line defined by 4 points
            sage: m=l.plot3d(z=2)
            sage: m.texture.opacity
            0.5
            sage: m.thickness
            10
            sage: L = line([(1,1), (1,2), (2,2), (2,1)], linestyle=':')
            sage: L.plot3d()
            Traceback (most recent call last):
            ...
            NotImplementedError: invalid 3d line style: ':'
        """
        if options is None:
            options = dict(self.options())
        options_3d = {}
        if 'thickness' in options:
            options_3d['thickness'] = options['thickness']
            del options['thickness']
        if 'zorder' in options:
            del options['zorder']
        if 'linestyle' in options:
            if options['linestyle'] not in ('-', 'solid'):
                raise NotImplementedError("invalid 3d line style: '%s'" %
                                          (options['linestyle']))
            del options['linestyle']
        options_3d.update(GraphicPrimitive_xydata._plot3d_options(self, options))
        return options_3d

    def plot3d(self, z=0, **kwds):
        """
        Plots a 2D line in 3D, with default height zero.

        EXAMPLES::

            sage: E = EllipticCurve('37a').plot(thickness=5).plot3d()                   # needs sage.schemes
            sage: F = EllipticCurve('37a').plot(thickness=5).plot3d(z=2)                # needs sage.schemes
            sage: E + F                         # long time (5s on sage.math, 2012), needs sage.schemes
            Graphics3d Object

        .. PLOT::

            E = EllipticCurve('37a').plot(thickness=5).plot3d()
            F = EllipticCurve('37a').plot(thickness=5).plot3d(z=2)
            sphinx_plot(E+F)
        """
        from sage.plot.plot3d.shapes2 import line3d
        options = self._plot3d_options()
        options.update(kwds)
        return line3d([(x, y, z) for x, y in zip(self.xdata, self.ydata)], **options)

    def _repr_(self):
        """
        String representation of a line primitive.

        EXAMPLES::

            sage: from sage.plot.line import Line
            sage: Line([-1,2,3,3], [17,4,0,2], {})._repr_()
            'Line defined by 4 points'
        """
        return "Line defined by %s points" % len(self)

    def __getitem__(self, i):
        """
        Extract the i-th element of the line (which is stored as a list of points).

        INPUT:

        - ``i`` -- integer between 0 and the number of points minus 1

        OUTPUT: 2-tuple of floats

        EXAMPLES::

            sage: L = line([(1,2), (3,-4), (2, 5), (1,2)])
            sage: line_primitive = L[0]; line_primitive
            Line defined by 4 points
            sage: line_primitive[0]
            (1.0, 2.0)
            sage: line_primitive[2]
            (2.0, 5.0)
            sage: list(line_primitive)
            [(1.0, 2.0), (3.0, -4.0), (2.0, 5.0), (1.0, 2.0)]
        """
        return self.xdata[i], self.ydata[i]

    def __setitem__(self, i, point):
        """
        Set the i-th element of this line (really a sequence of lines
        through given points).

        INPUT:

        - ``i`` -- integer between 0 and the number of points on the
          line minus 1

        - ``point`` -- a 2-tuple of floats

        EXAMPLES:

        We create a line graphics object ``L`` and get ahold of the
        corresponding line graphics primitive::

            sage: L = line([(1,2), (3,-4), (2, 5), (1,2)])
            sage: line_primitive = L[0]; line_primitive
            Line defined by 4 points

        We then set the 0th point to `(0,0)` instead of `(1,2)`::

            sage: line_primitive[0] = (0,0)
            sage: line_primitive[0]
            (0.0, 0.0)

        Plotting we visibly see the change -- now the line starts at `(0,0)`::

            sage: L
            Graphics object consisting of 1 graphics primitive
        """
        self.xdata[i] = float(point[0])
        self.ydata[i] = float(point[1])

    def __len__(self):
        r"""
        Return the number of points on this line (where a line is really a sequence
        of line segments through a given list of points).

        EXAMPLES:

        We create a line, then grab the line primitive as ``L[0]`` and compute
        its length::

            sage: L = line([(1,2), (3,-4), (2, 5), (1,2)])
            sage: len(L[0])
            4
        """
        return len(self.xdata)

    def _render_on_subplot(self, subplot):
        """
        Render this line on a matplotlib subplot.

        INPUT:

        - ``subplot`` -- a matplotlib subplot

        EXAMPLES:

        This implicitly calls this function::

            sage: line([(1,2), (3,-4), (2, 5), (1,2)])
            Graphics object consisting of 1 graphics primitive
        """
        from matplotlib import lines
        options = dict(self.options())
        for o in ('alpha', 'legend_color', 'legend_label', 'linestyle',
                  'rgbcolor', 'thickness'):
            if o in options:
                del options[o]
        p = lines.Line2D(self.xdata, self.ydata, **options)
        options = self.options()
        a = float(options['alpha'])
        p.set_alpha(a)
        p.set_linewidth(float(options['thickness']))
        p.set_color(to_mpl_color(options['rgbcolor']))
        p.set_label(options['legend_label'])
        # we don't pass linestyle in directly since the drawstyles aren't
        # pulled off automatically.  This (I think) is a bug in matplotlib 1.0.1
        if 'linestyle' in options:
            from sage.plot.misc import get_matplotlib_linestyle
            p.set_linestyle(get_matplotlib_linestyle(options['linestyle'],
                                                     return_type='short'))
        subplot.add_line(p)


def line(points, **kwds):
    """
    Return either a 2-dimensional or 3-dimensional line depending
    on value of points.

    INPUT:

    - ``points`` -- either a single point (as a tuple), a list of
      points, a single complex number, or a list of complex numbers

    For information regarding additional arguments, see either line2d?
    or line3d?.

    EXAMPLES::

        sage: line([(0,0), (1,1)])
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        E = line([(0,0), (1,1)])
        sphinx_plot(E)

    ::

        sage: line([(0,0,1), (1,1,1)])
        Graphics3d Object

    .. PLOT::

        E = line([(0,0,1), (1,1,1)])
        sphinx_plot(E)
    """
    try:
        return line2d(points, **kwds)
    except ValueError:
        from sage.plot.plot3d.shapes2 import line3d
        return line3d(points, **kwds)


@rename_keyword(color='rgbcolor')
@options(alpha=1, rgbcolor=(0, 0, 1), thickness=1, legend_label=None,
         legend_color=None, aspect_ratio='automatic')
def line2d(points, **options):
    r"""
    Create the line through the given list of points.

    INPUT:

    - ``points`` -- either a single point (as a tuple), a list of
      points, a single complex number, or a list of complex numbers

    Type ``line2d.options`` for a dictionary of the default options for
    lines.  You can change this to change the defaults for all future
    lines.  Use ``line2d.reset()`` to reset to the default options.

    INPUT:

    - ``alpha`` -- how transparent the line is

    - ``thickness`` -- how thick the line is

    - ``rgbcolor`` -- the color as an RGB tuple

    - ``hue`` -- the color given as a hue

    - ``legend_color`` -- the color of the text in the legend

    - ``legend_label`` -- the label for this item in the legend


    Any MATPLOTLIB line option may also be passed in.  E.g.,

    - ``linestyle`` -- (default: ``'-'``) the style of the line, which is one of
       - ``'-'`` or ``'solid'``
       - ``'--'`` or ``'dashed'``
       - ``'-.'`` or ``'dash dot'``
       - ``':'`` or ``'dotted'``
       - ``"None"`` or ``" "`` or ``""`` (nothing)

       The linestyle can also be prefixed with a drawing style (e.g., ``'steps--'``)

       - ``'default'`` (connect the points with straight lines)
       - ``'steps'`` or ``'steps-pre'`` (step function; horizontal
         line is to the left of point)
       - ``'steps-mid'`` (step function; points are in the middle of
         horizontal lines)
       - ``'steps-post'`` (step function; horizontal line is to the
         right of point)

    - ``marker``  -- the style of the markers, which is one of
       - ``"None"`` or ``" "`` or ``""`` (nothing) -- default
       - ``","`` (pixel), ``"."`` (point)
       - ``"_"`` (horizontal line), ``"|"`` (vertical line)
       - ``"o"`` (circle), ``"p"`` (pentagon), ``"s"`` (square), ``"x"`` (x), ``"+"`` (plus), ``"*"`` (star)
       - ``"D"`` (diamond), ``"d"`` (thin diamond)
       - ``"H"`` (hexagon), ``"h"`` (alternative hexagon)
       - ``"<"`` (triangle left), ``">"`` (triangle right), ``"^"`` (triangle up), ``"v"`` (triangle down)
       - ``"1"`` (tri down), ``"2"`` (tri up), ``"3"`` (tri left), ``"4"`` (tri right)
       - ``0`` (tick left), ``1`` (tick right), ``2`` (tick up), ``3`` (tick down)
       - ``4`` (caret left), ``5`` (caret right), ``6`` (caret up), ``7`` (caret down)
       - ``"$...$"`` (math TeX string)

    - ``markersize`` -- the size of the marker in points

    - ``markeredgecolor`` -- the color of the marker edge

    - ``markerfacecolor`` -- the color of the marker face

    - ``markeredgewidth`` -- the size of the marker edge in points

    EXAMPLES:

    A line with no points or one point::

        sage: line([])      # returns an empty plot
        Graphics object consisting of 0 graphics primitives
        sage: import numpy; line(numpy.array([]))                                       # needs numpy
        Graphics object consisting of 0 graphics primitives
        sage: line([(1,1)])
        Graphics object consisting of 1 graphics primitive

    A line with numpy arrays::

        sage: line(numpy.array([[1,2], [3,4]]))                                         # needs numpy
        Graphics object consisting of 1 graphics primitive

    A line with a legend::

        sage: line([(0,0),(1,1)], legend_label='line')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(line([(0,0),(1,1)], legend_label='line'))

    Lines with different colors in the legend text::

        sage: p1 = line([(0,0),(1,1)], legend_label='line')
        sage: p2 = line([(1,1),(2,4)], legend_label='squared', legend_color='red')
        sage: p1 + p2
        Graphics object consisting of 2 graphics primitives

    .. PLOT::

        p1 = line([(0,0),(1,1)], legend_label='line')
        p2 = line([(1,1),(2,4)], legend_label='squared', legend_color='red')
        sphinx_plot(p1 + p2)

    Extra options will get passed on to show(), as long as they are valid::

        sage: line([(0,1), (3,4)], figsize=[10, 2])
        Graphics object consisting of 1 graphics primitive
        sage: line([(0,1), (3,4)]).show(figsize=[10, 2]) # These are equivalent

    We can also use a logarithmic scale if the data will support it::

        sage: line([(1,2),(2,4),(3,4),(4,8),(4.5,32)],scale='loglog',base=2)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(line([(1,2),(2,4),(3,4),(4,8),(4.5,32)],scale='loglog',base=2))

    Many more examples below!

    A blue conchoid of Nicomedes::

        sage: from math import pi
        sage: L = [[1 + 5*cos(pi/2+pi*i/100),
        ....:       tan(pi/2+pi*i/100) * (1+5*cos(pi/2+pi*i/100))] for i in range(1,100)]
        sage: line(L, rgbcolor=(1/4,1/8,3/4))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        L = [[1+5*cos(pi/2+pi*i/100), tan(pi/2+pi*i/100)*(1+5*cos(pi/2+pi*i/100))] for i in range(1,100)]
        sphinx_plot(line(L, rgbcolor=(1/4,1/8,3/4)))

    A line with 2 complex points::

        sage: i = CC(0,1)
        sage: line([1 + i, 2 + 3*i])
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        i = CC(0,1)
        o = line([1+i, 2+3*i])
        sphinx_plot(o)

    A blue hypotrochoid (3 leaves)::

        sage: n = 4; h = 3; b = 2
        sage: L = [[n*cos(pi*i/100) + h*cos((n/b)*pi*i/100),
        ....:       n*sin(pi*i/100) - h*sin((n/b)*pi*i/100)] for i in range(200)]
        sage: line(L, rgbcolor=(1/4,1/4,3/4))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        n = 4
        h = 3
        b = 2
        L = [[n*cos(pi*i/100)+h*cos((n/b)*pi*i/100),n*sin(pi*i/100)-h*sin((n/b)*pi*i/100)] for i in range(200)]
        o = line(L, rgbcolor=(1/4,1/4,3/4))
        sphinx_plot(o)

    A blue hypotrochoid (4 leaves)::

        sage: n = 6; h = 5; b = 2
        sage: L = [[n*cos(pi*i/100) + h*cos((n/b)*pi*i/100),
        ....:       n*sin(pi*i/100) - h*sin((n/b)*pi*i/100)] for i in range(200)]
        sage: line(L, rgbcolor=(1/4,1/4,3/4))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        L = [[sin(pi*i/100)+sin(pi*i/50),-(1+cos(pi*i/100)+cos(pi*i/50))] for i in range(-100,101)]
        sphinx_plot(line(L, rgbcolor=(1,1/4,1/2)))

    A red limacon of Pascal::

        sage: L = [[sin(pi*i/100) + sin(pi*i/50),
        ....:       -(1 + cos(pi*i/100) + cos(pi*i/50))] for i in range(-100,101)]
        sage: line(L, rgbcolor=(1,1/4,1/2))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        L = [[sin(pi*i/100)+sin(pi*i/50),-(1+cos(pi*i/100)+cos(pi*i/50))] for i in range(-100,101)]
        sphinx_plot(line(L, rgbcolor=(1,1/4,1/2)))

    A light green trisectrix of Maclaurin::

        sage: L = [[2 * (1-4*cos(-pi/2+pi*i/100)^2),
        ....:       10 * tan(-pi/2+pi*i/100) * (1-4*cos(-pi/2+pi*i/100)^2)] for i in range(1,100)]
        sage: line(L, rgbcolor=(1/4,1,1/8))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        L = [[2*(1-4*cos(-pi/2+pi*i/100)**2),10*tan(-pi/2+pi*i/100)*(1-4*cos(-pi/2+pi*i/100)**2)] for i in range(1,100)]
        sphinx_plot(line(L, rgbcolor=(1/4,1,1/8)))

    A green lemniscate of Bernoulli::

        sage: cosines = [cos(-pi/2+pi*i/100) for i in range(201)]
        sage: v = [(1/c, tan(-pi/2+pi*i/100)) for i,c in enumerate(cosines) if c != 0]
        sage: L = [(a/(a^2+b^2), b/(a^2+b^2)) for a,b in v]
        sage: line(L, rgbcolor=(1/4,3/4,1/8))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        cosines = [cos(-pi/2+pi*i/100) for i in range(201)]
        v = [(1/c, tan(-pi/2+pi*i/100)) for i,c in enumerate(cosines) if c != 0]
        L = [(a/(a**2+b**2), b/(a**2+b**2)) for a,b in v]
        sphinx_plot(line(L, rgbcolor=(1/4,3/4,1/8)))

    A red plot of the Jacobi elliptic function `\text{sn}(x,2)`, `-3 < x < 3`::

        sage: L = [(i/100.0, real_part(jacobi('sn', i/100.0, 2.0)))
        ....:      for i in range(-300, 300, 30)]
        sage: line(L, rgbcolor=(3/4, 1/4, 1/8))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        L = [(i/100.0, real_part(jacobi('sn', i/100.0, 2.0))) for i in range(-300, 300, 30)]
        sphinx_plot(line(L, rgbcolor=(3/4, 1/4, 1/8)))

    A red plot of `J`-Bessel function `J_2(x)`, `0 < x < 10`::

        sage: L = [(i/10.0, bessel_J(2,i/10.0)) for i in range(100)]
        sage: line(L, rgbcolor=(3/4, 1/4, 5/8))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        L = [(i/10.0, bessel_J(2,i/10.0)) for i in range(100)]
        sphinx_plot(line(L, rgbcolor=(3/4,1/4,5/8)))

    A purple plot of the Riemann zeta function `\zeta(1/2 + it)`, `0 < t < 30`::

        sage: # needs sage.libs.pari sage.rings.complex_double
        sage: i = CDF.gen()
        sage: v = [zeta(0.5 + n/10 * i) for n in range(300)]
        sage: L = [(z.real(), z.imag()) for z in v]
        sage: line(L, rgbcolor=(3/4, 1/2, 5/8))
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        i = CDF.gen()
        v = [zeta(0.5 + n/10 * i) for n in range(300)]
        L = [(z.real(), z.imag()) for z in v]
        sphinx_plot(line(L, rgbcolor=(3/4,1/2,5/8)))

    A purple plot of the Hasse-Weil `L`-function `L(E, 1 + it)`, `-1 < t < 10`::

        sage: # needs sage.schemes
        sage: E = EllipticCurve('37a')
        sage: vals = E.lseries().values_along_line(1-I, 1+10*I, 100)  # critical line
        sage: L = [(z[1].real(), z[1].imag()) for z in vals]
        sage: line(L, rgbcolor=(3/4,1/2,5/8))
        Graphics object consisting of 1 graphics primitive

    .. PLOT ::

        E = EllipticCurve('37a')
        vals = E.lseries().values_along_line(1-I, 1+10*I, 100) # critical line
        L = [(z[1].real(), z[1].imag()) for z in vals]
        sphinx_plot(line(L, rgbcolor=(3/4,1/2,5/8)))

    A red, blue, and green "cool cat"::

        sage: # needs sage.symbolic
        sage: G = plot(-cos(x), -2, 2, thickness=5, rgbcolor=(0.5,1,0.5))
        sage: P = polygon([[1,2], [5,6], [5,0]], rgbcolor=(1,0,0))
        sage: Q = polygon([(-x,y) for x,y in P[0]], rgbcolor=(0,0,1))
        sage: G + P + Q   # show the plot
        Graphics object consisting of 3 graphics primitives

    .. PLOT::

        G = plot(-cos(x), -2, 2, thickness=5, rgbcolor=(0.5,1,0.5))
        P = polygon([[1,2], [5,6], [5,0]], rgbcolor=(1,0,0))
        Q = polygon([(-x,y) for x,y in P[0]], rgbcolor=(0,0,1))
        sphinx_plot(G + P + Q)

    TESTS:

    Check that :issue:`13690` is fixed. The legend label should have circles
    as markers.::

        sage: line(enumerate(range(2)), marker='o', legend_label='circle')
        Graphics object consisting of 1 graphics primitive
    """
    from sage.plot.graphics import Graphics
    from sage.plot.plot import xydata_from_point_list
    points = list(points)  # make sure points is a python list
    if not points:
        return Graphics()
    xdata, ydata = xydata_from_point_list(points)
    g = Graphics()
    g._set_extra_kwds(Graphics._extract_kwds_for_show(options))
    g.add_primitive(Line(xdata, ydata, options))
    if options['legend_label']:
        g.legend(True)
        g._legend_colors = [options['legend_color']]
    return g



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
cdef extern from "gsl/gsl_sf_elljac.h":

  int  gsl_sf_elljac_e(double u, double m, double * sn, double * cn, double * dn)




# distutils: libraries = flint
# distutils: depends = flint/fmpz_mod.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void fmpz_mod_ctx_init(fmpz_mod_ctx_t ctx, const fmpz_t n) noexcept
    void fmpz_mod_ctx_clear(fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_ctx_set_modulus(fmpz_mod_ctx_t ctx, const fmpz_t n) noexcept
    void fmpz_mod_set_fmpz(fmpz_t a, const fmpz_t b, const fmpz_mod_ctx_t ctx) noexcept
    bint fmpz_mod_is_canonical(const fmpz_t a, const fmpz_mod_ctx_t ctx) noexcept
    bint fmpz_mod_is_one(const fmpz_t a, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_add(fmpz_t a, const fmpz_t b, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_add_fmpz(fmpz_t a, const fmpz_t b, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_add_ui(fmpz_t a, const fmpz_t b, ulong c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_add_si(fmpz_t a, const fmpz_t b, slong c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_sub(fmpz_t a, const fmpz_t b, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_sub_fmpz(fmpz_t a, const fmpz_t b, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_sub_ui(fmpz_t a, const fmpz_t b, ulong c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_sub_si(fmpz_t a, const fmpz_t b, slong c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_fmpz_sub(fmpz_t a, const fmpz_t b, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_ui_sub(fmpz_t a, ulong b, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_si_sub(fmpz_t a, slong b, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_neg(fmpz_t a, const fmpz_t b, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_mul(fmpz_t a, const fmpz_t b, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_inv(fmpz_t a, const fmpz_t b, const fmpz_mod_ctx_t ctx) noexcept
    int fmpz_mod_divides(fmpz_t a, const fmpz_t b, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_pow_ui(fmpz_t a, const fmpz_t b, ulong e, const fmpz_mod_ctx_t ctx) noexcept
    int fmpz_mod_pow_fmpz(fmpz_t a, const fmpz_t b, const fmpz_t e, const fmpz_mod_ctx_t ctx) noexcept
    void fmpz_mod_discrete_log_pohlig_hellman_init(fmpz_mod_discrete_log_pohlig_hellman_t L) noexcept
    void fmpz_mod_discrete_log_pohlig_hellman_clear(fmpz_mod_discrete_log_pohlig_hellman_t L) noexcept
    double fmpz_mod_discrete_log_pohlig_hellman_precompute_prime(fmpz_mod_discrete_log_pohlig_hellman_t L, const fmpz_t p) noexcept
    const fmpz * fmpz_mod_discrete_log_pohlig_hellman_primitive_root(fmpz_mod_discrete_log_pohlig_hellman_t L) noexcept
    void fmpz_mod_discrete_log_pohlig_hellman_run(fmpz_t x, const fmpz_mod_discrete_log_pohlig_hellman_t L, const fmpz_t y) noexcept
    int fmpz_next_smooth_prime(fmpz_t a, const fmpz_t b) noexcept



"""
Local Normal Form
"""
# ****************************************************************************
#       Copyright (C) 2007 William Stein and Jonathan Hanke
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import copy

from sage.rings.infinity import Infinity
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.arith.misc import GCD, valuation, is_prime


def find_entry_with_minimal_scale_at_prime(self, p):
    r"""
    Find the entry of the quadratic form with minimal scale at the
    prime `p`, preferring diagonal entries in case of a tie.

    (I.e.  If
    we write the quadratic form as a symmetric matrix `M`, then this
    entry ``M[i,j]`` has the minimal valuation at the prime `p`.)

    .. NOTE::

        This answer is independent of the kind of matrix (Gram or
        Hessian) associated to the form.

    INPUT:

    - ``p`` -- a prime number > 0

    OUTPUT: a pair of integers `\geq 0`

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 2, [6, 2, 20]); Q
        Quadratic form in 2 variables over Integer Ring with coefficients:
          [ 6 2 ]
          [ * 20 ]
        sage: Q.find_entry_with_minimal_scale_at_prime(2)
        (0, 1)
        sage: Q.find_entry_with_minimal_scale_at_prime(3)
        (1, 1)
        sage: Q.find_entry_with_minimal_scale_at_prime(5)
        (0, 0)
    """
    n = self.dim()
    min_val = Infinity
    ij_index = None
    val_2 = valuation(2, p)
    for d in range(n):           # d = difference j-i
        for e in range(n - d):   # e is the length of the diagonal with value d.

            # Compute the valuation of the entry
            if d == 0:
                tmp_val = valuation(self[e, e + d], p)
            else:
                tmp_val = valuation(self[e, e + d], p) - val_2

            # Check if it's any smaller than what we have
            if tmp_val < min_val:
                ij_index = (e, e + d)
                min_val = tmp_val

    # Return the result
    return ij_index


def local_normal_form(self, p):
    r"""
    Return a locally integrally equivalent quadratic form over
    the `p`-adic integers `\ZZ_p` which gives the Jordan decomposition.

    The Jordan components are written as sums of blocks of size `\leq 2`
    and are arranged by increasing scale, and then by increasing norm.
    This is equivalent to saying that we put the `1 \times 1` blocks before
    the `2 \times 2` blocks in each Jordan component.

    INPUT:

    - ``p`` -- a positive prime number

    OUTPUT: a quadratic form over `\ZZ`

    .. WARNING::

        Currently this only works for quadratic forms defined over `\ZZ`.

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 2, [10,4,1])
        sage: Q.local_normal_form(5)
        Quadratic form in 2 variables over Integer Ring with coefficients:
          [ 1 0 ]
          [ * 6 ]

    ::

        sage: Q.local_normal_form(3)
        Quadratic form in 2 variables over Integer Ring with coefficients:
          [ 10 0 ]
          [ * 15 ]

        sage: Q.local_normal_form(2)
        Quadratic form in 2 variables over Integer Ring with coefficients:
          [ 1 0 ]
          [ * 6 ]
    """
    # Sanity Checks
    if self.base_ring() != ZZ:
        raise NotImplementedError("this currently only works for quadratic forms defined over ZZ")
    if not (p >= 2 and is_prime(p)):
        raise TypeError("p is not a positive prime number")

    # Some useful local variables
    Q = self.parent()(self.base_ring(), self.dim(), self.coefficients())

    # Prepare the final form to return
    Q_Jordan = copy.deepcopy(self)
    Q_Jordan.__init__(self.base_ring(), 0)

    while Q.dim() > 0:
        n = Q.dim()

        # Step 1: Find the minimally p-divisible matrix entry, preferring diagonals
        # -------------------------------------------------------------------------
        (min_i, min_j) = Q.find_entry_with_minimal_scale_at_prime(p)
        if min_i == min_j:
            min_val = valuation(2 * Q[min_i, min_j], p)
        else:
            min_val = valuation(Q[min_i, min_j], p)

        # Error if we still haven't seen nonzero coefficients!
        if min_val == Infinity:
            raise RuntimeError("the original matrix is degenerate")

        # Step 2: Arrange for the upper leftmost entry to have minimal valuation
        # ----------------------------------------------------------------------
        if min_i == min_j:
            block_size = 1
            Q.swap_variables(0, min_i, in_place=True)
        else:
            # Work in the upper-left 2x2 block, and replace it by its 2-adic equivalent form
            Q.swap_variables(0, min_i, in_place=True)
            Q.swap_variables(1, min_j, in_place=True)

            # 1x1 => make upper left the smallest
            if p != 2:
                block_size = 1
                Q.add_symmetric(1, 0, 1, in_place=True)
            # 2x2 => replace it with the appropriate 2x2 matrix
            else:
                block_size = 2

        # Step 3: Clear out the remaining entries
        #  ---------------------------------------
        min_scale = p ** min_val                             # This is the minimal valuation of the Hessian matrix entries.

        # Perform cancellation over Z by ensuring divisibility
        if block_size == 1:
            a = 2 * Q[0, 0]
            for j in range(block_size, n):
                b = Q[0, j]
                g = GCD(a, b)

                # Sanity Check:  a/g is a p-unit
                if valuation(g, p) != valuation(a, p):
                    raise RuntimeError("we have a problem with our rescaling not preserving p-integrality")

                Q.multiply_variable(ZZ(a / g), j, in_place=True)   # Ensures that the new b entry is divisible by a
                Q.add_symmetric(ZZ(-b / g), j, 0, in_place=True)  # Performs the cancellation

        elif block_size == 2:
            a1 = 2 * Q[0, 0]
            a2 = Q[0, 1]
            b1 = Q[1, 0]      # This is the same as a2
            b2 = 2 * Q[1, 1]

            big_det = a1 * b2 - a2 * b1
            small_det = big_det / (min_scale * min_scale)

            # Cancels out the rows/columns of the 2x2 block
            for j in range(block_size, n):
                a = Q[0, j]
                b = Q[1, j]

                # Ensures an integral result (scale jth row/column by big_det)
                Q.multiply_variable(big_det, j, in_place=True)

                # Performs the cancellation (by producing -big_det * jth row/column)
                Q.add_symmetric(ZZ(-(a * b2 - b * a2)), j, 0, in_place=True)
                Q.add_symmetric(ZZ(-(-a * b1 + b * a1)), j, 1, in_place=True)

                # Now remove the extra factor (non p-unit factor) in big_det we introduced above
                Q.divide_variable(ZZ(min_scale * min_scale), j, in_place=True)

            # Uses Cassels's proof to replace the remaining 2 x 2 block
            if (1 + small_det) % 8 == 0:
                Q[0, 0] = 0
                Q[1, 1] = 0
                Q[0, 1] = min_scale
            elif (5 + small_det) % 8 == 0:
                Q[0, 0] = min_scale
                Q[1, 1] = min_scale
                Q[0, 1] = min_scale
            else:
                raise RuntimeError("Error in LocalNormal: Impossible behavior for a 2x2 block! \n")

        # Check that the cancellation worked, extract the upper-left block, and trim Q to handle the next block.
        for i in range(block_size):
            for j in range(block_size, n):
                if Q[i, j] != 0:
                    raise RuntimeError(f"the cancellation did not work properly at entry ({i},{j})")
        Q_Jordan = Q_Jordan + Q.extract_variables(range(block_size))
        Q = Q.extract_variables(range(block_size, n))

    return Q_Jordan


def jordan_blocks_by_scale_and_unimodular(self, p, safe_flag=True):
    r"""
    Return a list of pairs `(s_i, L_i)` where `L_i` is a maximal
    `p^{s_i}`-unimodular Jordan component which is further decomposed into
    block diagonals of block size `\le 2`.

    For each `L_i` the `2 \times 2` blocks are listed after the `1 \times 1` blocks
    (which follows from the convention of the
    :meth:`local_normal_form` method).

    .. NOTE::

        The decomposition of each `L_i` into smaller blocks is not unique!

    The ``safe_flag`` argument allows us to select whether we want a copy of
    the output, or the original output.  By default ``safe_flag = True``, so we
    return a copy of the cached information.  If this is set to ``False``, then
    the routine is much faster but the return values are vulnerable to being
    corrupted by the user.

    INPUT:

    - ``p`` -- a prime number > 0

    OUTPUT:

    A list of pairs `(s_i, L_i)` where:

    - `s_i` is an integer,
    - `L_i` is a block-diagonal unimodular quadratic form over `\ZZ_p`.

    .. NOTE::

        These forms `L_i` are defined over the `p`-adic integers, but by a
        matrix over `\ZZ` (or `\QQ`?).

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,9,5,7])
        sage: Q.jordan_blocks_by_scale_and_unimodular(3)
        [(0, Quadratic form in 3 variables over Integer Ring with coefficients:
               [ 1 0 0 ]
               [ * 5 0 ]
               [ * * 7 ]),
         (2, Quadratic form in 1 variables over Integer Ring with coefficients:
               [ 1 ])]

    ::

        sage: Q2 = QuadraticForm(ZZ, 2, [1,1,1])
        sage: Q2.jordan_blocks_by_scale_and_unimodular(2)
        [(-1, Quadratic form in 2 variables over Integer Ring with coefficients:
                [ 2 2 ]
                [ * 2 ])]
        sage: Q = Q2 + Q2.scale_by_factor(2)
        sage: Q.jordan_blocks_by_scale_and_unimodular(2)
        [(-1, Quadratic form in 2 variables over Integer Ring with coefficients:
                [ 2 2 ]
                [ * 2 ]),
         (0, Quadratic form in 2 variables over Integer Ring with coefficients:
                [ 2 2 ]
                [ * 2 ])]
    """
    # Try to use the cached result
    try:
        if safe_flag:
            return copy.deepcopy(self.__jordan_blocks_by_scale_and_unimodular_dict[p])
        return self.__jordan_blocks_by_scale_and_unimodular_dict[p]
    except (KeyError, AttributeError):
        # Initialize the global dictionary if it doesn't exist
        if not hasattr(self, '__jordan_blocks_by_scale_and_unimodular_dict'):
            self.__jordan_blocks_by_scale_and_unimodular_dict = {}

    # Deal with zero dim'l forms
    if self.dim() == 0:
        return []

    # Find the Local Normal form of Q at p
    Q1 = self.local_normal_form(p)

    # Parse this into Jordan Blocks
    n = Q1.dim()
    tmp_Jordan_list = []
    i = 0
    start_ind = 0
    if n >= 2 and Q1[0, 1] != 0:
        start_scale = valuation(Q1[0, 1], p) - 1
    else:
        start_scale = valuation(Q1[0, 0], p)

    while i < n:

        # Determine the size of the current block
        if i == n - 1 or Q1[i, i + 1] == 0:
            block_size = 1
        else:
            block_size = 2

        # Determine the valuation of the current block
        if block_size == 1:
            block_scale = valuation(Q1[i, i], p)
        else:
            block_scale = valuation(Q1[i, i + 1], p) - 1

        # Process the previous block if the valuation increased
        if block_scale > start_scale:
            tmp_Jordan_list += [(start_scale, Q1.extract_variables(range(start_ind, i)).scale_by_factor(ZZ.one() / QQ(p)**start_scale))]
            start_ind = i
            start_scale = block_scale

        # Increment the index
        i += block_size

    # Add the last block
    tmp_Jordan_list += [(start_scale, Q1.extract_variables(range(start_ind, n)).scale_by_factor(ZZ.one() / QQ(p)**start_scale))]

    # Cache the result
    self.__jordan_blocks_by_scale_and_unimodular_dict[p] = tmp_Jordan_list

    # Return the result
    return tmp_Jordan_list


def jordan_blocks_in_unimodular_list_by_scale_power(self, p):
    r"""
    Return a list of Jordan components, whose component at index `i`
    should be scaled by the factor `p^i`.

    This is only defined for integer-valued quadratic forms
    (i.e., forms with base ring `\ZZ`), and the indexing only works
    correctly for `p=2` when the form has an integer Gram matrix.

    INPUT:

    - ``self`` -- a quadratic form over `\ZZ`, which has integer Gram matrix if `p = 2`
    - ``p`` -- a prime number > 0

    OUTPUT: list of `p`-unimodular quadratic forms

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 3, [2, -2, 0, 3, -5, 4])
        sage: Q.jordan_blocks_in_unimodular_list_by_scale_power(2)
        Traceback (most recent call last):
        ...
        TypeError: the given quadratic form has a Jordan component with a negative scale exponent

        sage: Q.scale_by_factor(2).jordan_blocks_in_unimodular_list_by_scale_power(2)
        [Quadratic form in 2 variables over Integer Ring with coefficients:
           [ 0 2 ]
           [ * 0 ],
         Quadratic form in 0 variables over Integer Ring with coefficients:
           ,
         Quadratic form in 1 variables over Integer Ring with coefficients:
           [ 345 ]]

        sage: Q.jordan_blocks_in_unimodular_list_by_scale_power(3)
        [Quadratic form in 2 variables over Integer Ring with coefficients:
           [ 2 0 ]
           [ * 10 ],
         Quadratic form in 1 variables over Integer Ring with coefficients:
           [ 2 ]]
    """
    # Sanity Check
    if self.base_ring() != ZZ:
        raise TypeError("this method only makes sense for integer-valued quadratic forms (i.e. defined over ZZ)")

    # Deal with zero dim'l forms
    if self.dim() == 0:
        return []

    # Find the Jordan Decomposition
    list_of_jordan_pairs = self.jordan_blocks_by_scale_and_unimodular(p)
    scale_list = [P[0] for P in list_of_jordan_pairs]
    s_max = max(scale_list)
    if min(scale_list) < 0:
        raise TypeError("the given quadratic form has a Jordan component with a negative scale exponent")

    # Make the new list of unimodular Jordan components
    zero_form = self.parent()(ZZ, 0)
    list_by_scale = [zero_form for _ in range(s_max + 1)]
    for P in list_of_jordan_pairs:
        list_by_scale[P[0]] = P[1]

    # Return the new list
    return list_by_scale
